   
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  #######################################################################
  ####      Function of AICC Loss Score for Smoothing Bandwidth      ####
  #######################################################################
  ####        version 1.0 Created by Chong Liu on 2013/03/17         ####
  #######################################################################
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#                                              
  # Input
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.H.Log:          Log bandwidth
  # 2.Data:           Binned data set in the format of binning function in sm package
  # 3.Structure.2D:   Type of bandwidths for two dimensional data
  # 4.LOPO:           Artificial variable
  # 5.LocLinQua:      Logical value indicating whether local linear quadratic smoothing is used for two dimensional data
  # 6.LLQParams:      A list including
  #                   Method:     Methods used for local linear-quadratic smoothing
  #                               "poly" uses (1,X1-X0,X2-X0,(X2-X0)^2 and MA750 type of polynomial expressions to solve for the local fitting
  #                               "matrix" uses (1,X1,X2,X2^2) and compact matrix expressions to solve for the local fitting
  #                   InvSym:     A facas object in Ryacas package of matrix inverse used for corresponding method indicated by Method. Reason we need this argument is symbolic 
  #                               inverse of 4 by 4 matrix takes a while. In simulations, this symbolic inverse is the same for each simulated data but evaluated at different points. 
  #                               It is time consuming to do this inverse each time. Instead, it can be provided from outside and we only evaluate the inverse object at different points
  #                               using this function in simulations.
  # 7.DF:             Artificial variable
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Output
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Loss:             AICC score
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#


  AICCLocSmoothing <- function(H.Log, Data, Structure.2D, LOPO = NA, LocLinQua, LLQParams, DF = NA)
  {  
    
    # prepare variables
    
    H <- exp(H.Log) # input is log bandwidth
    X <- as.matrix(Data$x)
    Y <- as.matrix(Data$means) # Y is needed for CV calculation
    Weights <- Data$x.freq # weights derived from binning
    Devs <- Data$devs # Y deviation from the mean
    N <- dim(X)[1] # number of observations
    D <- dim(X)[2] # dimension of the independent data    
    
    # modify the bandwidth H
    
    if(D == 2)
    {
      if(Structure.2D == "common"){H <- c(H, H)}
      if(Structure.2D == "scaled"){H <- sd(X) * H}
    }    
    
    # get full sample smoothing and Hat matrix
    
    if(D == 2 & LocLinQua)
    {
      FitList <- LocLinQuaReg2D(X = X, Y = Y, X0 = X, H = H, Weights = Weights, NbinsX = 0, NbinsX0 = 0, Method = LLQParams$Method, InvSym = LLQParams$InvSym)      
    }else
    {
      FitList <- LocLinReg(X = X, Y = Y, X0 = X, H = H, Weights = Weights, NbinsX = 0, NbinsX0 = 0)  
    }      
    Hat <- FitList$H
    M <- FitList$M
    
    # compute AICC loss
    
    HatWhtTrace <- sum(Weights * diag(Hat))
    NRaw <- sum(Weights)
    SigSq <- sum(Weights * ((diag(N) - Hat) %*% Y)^2 + Devs) / NRaw
    Penalty <- 1 + 2 * (HatWhtTrace + 1)/(NRaw - HatWhtTrace - 2)
    Loss <- log(SigSq) + Penalty
    if(is.na(Loss)) 
    {
      Loss <- 1e+10
    }
    
    # return loss
    
    return(Loss)	
  }