  
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  #######################################################################
  ####    Function of Fitting Anisotropy Matern Correlation Model    ####
  #######################################################################
  ####        version 1.0 Created by Chong Liu on 2012/11/13         ####
  #######################################################################
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#                                              
  # Input
  # 1.Estimates:          A vector of length NTarget, which is the estimate correlation corresponding to each spatial separation vector
  # 2.DefaultPars:        A matrix of size NTask by 4. Each row is default values for variables not to be optimized over. 
  #                       NA values in the matrix suggests variables to be optimized.
  # 3.InitialPars:        A matrix of size NTask by 4. Each row is initial values for variables to be optimized. 
  #                       NA values suggest variables not to be optimized over. Not available for "Grid"
  # 4.ParsBound:          A array of size NTask by 2 by 4. The second dimension corresponds to lower bound and upper bound. The third dimension corresponds to 4 parameters.
  #                       NA values suggest variables not be optimized over. Not available for "Nelder-Mead" and "BFGS"
  # 5.TargetRep:          A matrix of size of NTarget by 2, which are spatial separation vectors corresponding to estimated correlations
  # 6.TargetType:         The type of spatial separation vector in TargetRep
  #                       "cartesian"
  #                       "polar"
  # 7.Optim.Method:       A vector of length NTask, where each entry is the name of optimization method to be used for each task.
  #                       "Nelder-Mead", "BFGS", "L-BFGS-B", "nls", "Grid"
  # 8.Log:                A matrix of size NTask by 4, indicating whether each variable has been taken logarithm
  # 9.MaxIt:              A vector of length NTask, specifies number of iterations for each task, not available for "Grid"
  # 10.NGridsSearch:      Only for "Grid", a matrix of size NTask by 4, specifies number of grids for each task and each variable 
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Output
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.ParFitted:          A matrix of size NTask by 4
  # 2.Converge:           A vector of length NTask, not available for "Grid"
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#

  AnisoMaternFit <- function(Estimates, DefaultPars, InitialPars = NULL, ParsBound = NULL, TargetRep, TargetType, Optim.Method, Log, 
                             MaxIt = NULL, NGridsSearch = NULL)
  {
    # include packages and functions
    
      #source("MaternLossWrap.r")
      #source("AnisoMaternWrap.r")
    
    # check consistency
    
      NMethod <- length(Optim.Method)
      NDefault <- dim(DefaultPars)[1]
      if(!is.null(InitialPars)){NInitial <- dim(InitialPars)[1]}else{NInitial <- NMethod}
      if(!is.null(ParsBound)){NBound <- dim(ParsBound)[1]}else{NBound <- NMethod}      
      if(!is.null(MaxIt)){NMaxIt <- length(MaxIt)}else{NMaxIt <- NMethod}
      if(!all(c(NMethod,NDefault,NInitial,NBound,NMaxIt)-rep(NMethod,5) == 0)){stop("Task numbers are not consistent")}      
      
    # perform model fitting
      
      NTask <- length(Optim.Method)
      #Optim.Method <- tolower(Optim.Method)
      ParFitted <- matrix(NA, NTask, 4)
      Converge <- c()
      for(task in 1 : NTask)
      {
        thisOptim.Method <- Optim.Method[task]
        thisAlpha <- DefaultPars[task, 1]
        thisDelta <- DefaultPars[task, 2]
        thisKappa <- DefaultPars[task, 3]
        thisPhi <- DefaultPars[task, 4]
        thisLog <- as.logical(Log[task,])
            
        if(tolower(thisOptim.Method) == "nelder-mead" | tolower(thisOptim.Method) == "bfgs")
        {
          thisInitial <- as.vector(na.omit(InitialPars[task,]))
          thisMaxIt <- MaxIt[task]
          thisFit <- optim(par = thisInitial, fn = MaternLossWrap, Alpha = thisAlpha, Delta = thisDelta, Kappa = thisKappa, Phi = thisPhi, 
                           SpaSep = TargetRep, SepType = TargetType, Estimates = Estimates, method = thisOptim.Method, Log = thisLog, 
                           control = list(maxit = thisMaxIt))
          ParFitted[task,] <- DefaultPars[task,]
          ParFitted[task,is.na(ParFitted[task,])] <- thisFit$par
          Converge[task] <- thisFit$convergence
        }
        if(tolower(thisOptim.Method) == "l-bfgs-b")
        {
          thisInitial <- as.vector(na.omit(InitialPars[task,]))
          thisLowerBound <- as.vector(na.omit(ParsBound[task,1,]))
          thisUpperBound <- as.vector(na.omit(ParsBound[task,2,]))
          thisMaxIt <- MaxIt[task]
          thisFit <- optim(par = thisInitial, fn = MaternLossWrap, Alpha = thisAlpha, Delta = thisDelta, Kappa = thisKappa, Phi = thisPhi, 
                           SpaSep = TargetRep, SepType = TargetType, Estimates = Estimates, method = thisOptim.Method, Log = thisLog,
                           lower = thisLowerBound, upper = thisUpperBound, control = list(maxit = thisMaxIt))
          ParFitted[task,] <- DefaultPars[task,]
          ParFitted[task,is.na(ParFitted[task,])] <- thisFit$par
          Converge[task] <- thisFit$convergence
        }       
        if(tolower(thisOptim.Method) == "nls")
        { 
          thisInitial <- as.vector(na.omit(InitialPars[task,]))
          thisLowerBound <- as.vector(na.omit(ParsBound[task,1,]))
          thisUpperBound <- as.vector(na.omit(ParsBound[task,2,]))
          thisMaxIt <- MaxIt[task]
          thisFit <- nls(Estimates ~ AnisoMaternWrap(Parameters, Alpha = thisAlpha, Delta = thisDelta, Kappa = thisKappa, Phi = thisPhi, SpaSep = TargetRep, SepType = TargetType, Log = thisLog), start = list(Parameters = thisInitial), algorithm = "port", 
                         lower = thisLowerBound, upper = thisUpperBound, control = list(maxiter = thisMaxIt, warnOnly = TRUE))
          ParFitted[task,] <- DefaultPars[task,]
          ParFitted[task,is.na(ParFitted[task,])] <- summary(thisFit)$coefficients[,1]
          Converge[task] <- thisFit$convergence          
        }
        if(tolower(thisOptim.Method) == "grid")
        {
          thisLowerBound <- ParsBound[task,1,]
          thisLowerBound[is.na(thisLowerBound)] <- as.vector(na.omit(DefaultPars[task,]))
          thisUpperBound <- ParsBound[task,2,]
          thisUpperBound[is.na(thisUpperBound)] <- as.vector(na.omit(DefaultPars[task,]))
          thisNGrids <- NGridsSearch[task,]
          AlphaVec <- seq(thisLowerBound[1], thisUpperBound[1], length.out = thisNGrids[1])
          DeltaVec <- seq(thisLowerBound[2], thisUpperBound[2], length.out = thisNGrids[2])
          KappaVec <- seq(thisLowerBound[3], thisUpperBound[3], length.out = thisNGrids[3])
          PhiVec <- seq(thisLowerBound[4], thisUpperBound[4], length.out = thisNGrids[4])
          Loss <- array(NA, dim = thisNGrids)
          for(aa in 1 : thisNGrids[1])
          {
            for(dd in 1 : thisNGrids[2])
            {
              for(kk in 1 : thisNGrids[3])
              {
                for(pp in 1 : thisNGrids[4])
                {
                  AlphaNow <- AlphaVec[aa]
                  DeltaNow <- DeltaVec[dd]
                  KappaNow <- KappaVec[kk]
                  PhiNow <- PhiVec[pp]
                  Loss[aa, dd, kk, pp] <- MaternLoss(Alpha = AlphaNow, Delta = DeltaNow, Kappa = KappaNow, Phi = PhiNow, SpaSep = TargetRep, SepType = TargetType, Log = thisLog, Estimates = Estimates)
                }
              }
            }
          }
          temp <- which(Loss == min(Loss), arr.ind = TRUE)[1,]
          ParFitted[task,] <- c(AlphaVec[temp[1]], DeltaVec[temp[2]], KappaVec[temp[3]], PhiVec[temp[4]])
          Converge[task] <- NA
        }
        if(Log[task, 1]){ParFitted[task, 1] <- exp(ParFitted[task, 1])} 
        if(Log[task, 2]){ParFitted[task, 2] <- exp(ParFitted[task, 2])} 
        if(Log[task, 3]){ParFitted[task, 3] <- exp(ParFitted[task, 3])} 
        if(Log[task, 4]){ParFitted[task, 4] <- exp(ParFitted[task, 4])}         
      }
      
    # return resutls
      
      return(list(ParFitted = ParFitted, Converge = Converge))
  }