  
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  ##########################################################################
  ####    Function to Generate Training and Test Set for SCorEstMMCV    ####
  ##########################################################################
  ####          version 1.0 Created by Chong Liu on 2012/11/14          ####
  ##########################################################################
  ####                                                                  ####
  ##########################################################################
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#                                              
  # Input:
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.Dim:              The dimension of the curve locations. 
  #                     If dim = 1, then all curves are aligned on a 1 dimensional line. If dim = 2, then curves are located at a 2 dimensional rectangular space
  # 2.Coordinate:       The cartesian coordinates of the curve locations
  # 3.D1NFold:          Number of folds used for 1 dimensional locations
  # 4.D1Separation:     The smallest distance between testing locations and training locations used for 1 dimensional locations
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Output
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.Train:            A list, each cell of which contains curve IDs, corresponding to a fold of training curves
  # 2.Test:             A list, each cell of which contains curve IDs, corresponding to a fold of testing curves
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Notes
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # For 1 dimensional locations, D1NFold and D1Separation are needed to generate testing and training sets
  # For 2 dimensional locations, current strategy is to generate 4 testing sets which correspond to the 4 corners of the rectangular space 
  # For each testing set on the corner, the training set is the corner on the diagonal
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  CVTrainTestGen <- function(Dim, Coordinate, D1NFold = NA, D1Separation = NA)
  {
    # include packages and functions
    
      #source("U.RepMat.r")
    
    # construct training set and test set for 1 dimensional locations
      
      if(Dim == 1)
      {
        Train <- list()
        Test <- list()
        DistMat <- DistGen(Coordinate = Coordinate, DistType = "euclidean", Alpha = 0, Delta = 1)
        NCurve <- dim(DistMat)[1]
        for(i in 1 : D1NFold)
        {
          FoldSize <- floor(NCurve / D1NFold)
          Test[[i]] <- ((i - 1) * FoldSize + 1) : (i * FoldSize)
          tempDist <- DistMat[Test[[i]],]
          Train[[i]] <- which(apply(tempDist > D1Separation, MARGIN = 2, FUN = all))
        }
      }
      
    # construct training set and test set for 2 dimensional locations ( rectangle )
      
      if(Dim == 2)
      {
        Train <- list()
        Test <- list()
        coor.x <- Coordinate[,1]
        coor.y <- Coordinate[,2]
        Edge <- sqrt(length(coor.x)[1])
        PatchEdge <- floor(Edge / 2)
        Train[[1]] <- which(coor.x >= 1 & coor.x <= PatchEdge & coor.y <= Edge & coor.y >= (Edge-PatchEdge+1))
        Train[[2]] <- which(coor.x >= (Edge-PatchEdge+1) & coor.x <= Edge & coor.y <= Edge & coor.y >= (Edge-PatchEdge+1))
        Train[[3]] <- which(coor.x >= 1 & coor.x <= PatchEdge & coor.y <= PatchEdge & coor.y >= 1)
        Train[[4]] <- which(coor.x >= (Edge-PatchEdge+1) & coor.x <= Edge & coor.y <= PatchEdge & coor.y >= 1)
        Test[[1]] <- Train[[4]]
        Test[[2]] <- Train[[3]]
        Test[[3]] <- Train[[2]]
        Test[[4]] <- Train[[1]]        
      }
      
    # return results
      
      list(Train = Train, Test = Test)
  }