  
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  ##################################################################
  ####       Function to Transform Data into List Format        ####
  ##################################################################
  ####      version 1.0 Created by Chong Liu on 2012/10/30      ####
  ##################################################################
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#                                              
  # Input:
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.Y:              Observations of underlying random function.
  #                   @ Matrix format indicates Y is observed at regular grids of time points, however, maybe with missingness. Rows correspond to locations and columns correspond 
  #                   to time points.
  #                   @ List format indicates Y is observed at irregular time points. The list is of length Ncurve in which each cell stores the observations on each individual curve
  # 2.Time:           Time points corresponding to Y, in the same format as Y. 
  # 3.TimeRange:      The time range of Time, used to normalize Time onto [0,1] interval.
  # 4.Format:         Type of results
  #                   "Stack" indicates that the results are stacked together cross all curves;
  #                   "PerCurve" indicates that the results are in list format where each cell stores information of that individual curve.
  # 5.TimeRegular:    A logical variable indicating whether observation times shoudl be coerced to the near regular time points according to EvalTime.
  # 6.EvalTime:       Evaluation times used to approximate real time.
  # 7.Rounding:       Way of approximating real time using EvalTime
  #                   "up" approximates each real time point to the nearest EvalTime from above;
  #                   "down" approximates each real time point to the nearest EvalTime from below;
  #                   "cloest" approximates each real time point to the nearest EvalTime from either side; 
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Output
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.DataList:        A list of length 5. If format is "stack", then each cell is a single vector of variable from all observations stacked together. If the format is "PerCurve" , 
  #                   each cell is also a list, each cell of which contains observations on an individual curve
  #                   The first cell, "ID", is curve IDs in stack format indicating which curve each observation or time corresponds to;
  #                   The second cell, "Y", is the transformed Y in either"Stack" or "PerCurve" format;
  #                   The third cell, "Time", is the transformed time in either "Stack" or "PerCurve" format;
  #                   The fourth cell, "TimeApprox", is the EvalTime approximation to the real Time in either "Stack" or "PerCurve" format;
  #                   The fifth cell, "TimeApproxIdx", is the index of real Time approximated by EvalTime, in either "Stack" or "PerCurve" format.
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Details
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # If the input Y and Time are in list format and output type is "PerCurve", then this function does NOT do any transformation.
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  

	DataMassage <- function(Y, Time, TimeRange, Format = "Stack", TimeRegular = FALSE, EvalTime = NULL, Rounding = "closest")
	{
    
    # include packages and functions
    
    # check input data consistency
      
      if(class(Y) != class(Time)){stop("Y and Time are not consistent")} # Y and Time must be in the same format, either matrix or list.
      if(TimeRegular & is.null(EvalTime)){stop("Evaltime must be provided if TimeRegular is TRUE")} # if forcing Time to be regular, the regular time grids EvalTime must be provided
    
		# transform data on regular grids from matrix to list

			if(is.matrix(Y))
			{
				NCurve <- nrow(Y)
				NoMissYList <- lapply(split(Y, 1 : NCurve), FUN = function(x){as.vector(na.omit(x))}) # extract nonmissing values for each curve
				names(NoMissYList) <- NULL # remove wierd names
				NoMissTimeList <- lapply(split(Time, 1 : NCurve), FUN = function(x){as.vector(na.omit(x))}) # extract nonmissing times for each curve
				names(NoMissTimeList) <- NULL # remove wierd names
        Y <- NoMissYList # assign Y as the list version without missingness
        Time <- NoMissTimeList # assign Time as the list version without missingness
			}else
			{
        NCurve <- length(Y)
			}
      
		# transform Time onto interval [0,1]

		  Time <- lapply(Time, FUN = function(x){(x - TimeRange[1]) / (TimeRange[2] - TimeRange[1])})
      
    # force Time into regular if needed
      
      if(TimeRegular)
      {
        EvalTime <- (EvalTime - TimeRange[1]) / (TimeRange[2] - TimeRange[1]) # transform EvalTime onto interval [0,1]
        tempfun <- function(x){outer(x, EvalTime, FUN = "-")} # function to compute distance of time points to each evaluation time points
        Rounding <- tolower(Rounding)
        if(Rounding == "down"){TimeApproxIdx <- lapply(Time, FUN = function(x){y<-tempfun(x);z<-apply(y, MARGIN = 1, FUN = function(y){which(y == min(y[y >= 0]))})})} # or use cut(x, breaks = EvalTime, labels = FALSE)
        if(Rounding == "up"){TimeApproxIdx <- lapply(Time, FUN = function(x){y<-tempfun(x);z<-apply(y, MARGIN = 1, FUN = function(y){which(y == max(y[y <= 0]))})})} # or use cut(x, breaks = EvalTime, labels = FALSE) + 1
        if(Rounding == "closest"){TimeApproxIdx <- lapply(Time, FUN = function(x){y<-tempfun(x);z<-apply(y, MARGIN = 1, FUN = function(y){which(abs(y) == min(abs(y)))})})}
        TimeApprox <- lapply(TimeApproxIdx, FUN = function(x){EvalTime[x]}) # new time approximated by the EvalTime
      }else
      {
        TimeApproxIdx <- NULL
        TimeApprox <- NULL
      }
      
    # generate ID after everything is in list format
      
      ID <- rep(1 : NCurve, sapply(Y, FUN = length))
      
		# massage list data into desired structure 
      
      Format <- tolower(Format);
			if(Format == "stack")
			{
				Y <- unlist(Y)
				Time <- unlist(Time)
				TimeApprox <- unlist(TimeApprox)
				TimeApproxIdx <- unlist(TimeApproxIdx)
				DataList <- list(ID = ID, Y = Y, Time = Time, TimeApprox = TimeApprox, TimeApproxIdx = TimeApproxIdx)
													
			}
			if(Format == "percurve")
			{
        
				DataList <- list(Y = Y, Time = Time, TimeApprox = TimeApprox, TimeApproxIdx = TimeApproxIdx)
			}

    # return DataMat
      
		  return(DataList)
	}