  
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  ################################################################
  ####             Function of Selecting Bandwidth            ####
  ################################################################
  ####     version 1.0 Created by Chong Liu on 2013/02/28     ####
  ################################################################
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#                                              
  # Input
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.X:              Observed independent variable
  # 2.Y:              Observed dependent variable
  # 3.Structure.2D:   Type of bandwidths for two dimensional data
  # 4.BWMethod:       "cv", leave-one-point-out or leave-one-curve-out cross validation 
  #                   "df", sum of diagonal of the hat matrix
  #                   "aicc", modified aic method
  # 5.LOPO:           Way of leave one point out
  #                   LOPO = 0, NOT leave-one-curve-out, meaning leave-one-curve-out, still under construction
  #                   LOPO = 1, reduce weight by 1 
  #                   LOPO = 2, reduce weight to 0
  # 6.LocLinQua:      Logical value indicating whether local linear quadratic smoothing is used for two dimensional data
  # 7.NbinsX:         number of bins used for design matrix X
  #                   NbinsX = -1, automatically determine the number of bins used according to SM package
  #                   NbinsX = 0, no binning
  #                   NbinsX is a positive integer, number of bins used
  # 8.LLQParams:      A list of parameters for local linear quadratic smoothing, including
  #                   Method:     Methods used for local linear-quadratic smoothing
  #                               "poly" uses (1,X1-X0,X2-X0,(X2-X0)^2 and MA750 type of polynomial expressions to solve for the local fitting
  #                               "matrix" uses (1,X1,X2,X2^2) and compact matrix expressions to solve for the local fitting
  #                   InvSym:     A facas object in Ryacas package of matrix inverse used for corresponding method indicated by Method. Reason we need this argument is symbolic 
  #                               inverse of 4 by 4 matrix takes a while. In simulations, this symbolic inverse is the same for each simulated data but evaluated at different points. 
  #                               It is time consuming to do this inverse each time. Instead, it can be provided from outside and we only evaluate the inverse object at different points
  #                               using this function in simulations.
  # 9.DF:             Degree of freedom used in DF method in choosing bandwidth
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Output
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # BW:              Selected BW
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#

  HSelect <- function(X, Y, Structure.2D = NULL, BWMethod, LOPO, LocLinQua, NbinsX, LLQParams = NULL, DF = NA)
  {
    # check data consistency
    
    if(!is.matrix(X) | !is.matrix(Y)){stop("input must be in matrix format")}      
    
    # binning design matrix X
    
    D <- dim(X)[2] # dimension of regression  
    N <- dim(X)[1] # number of observations  
    if(NbinsX == -1){NbinsX <- round((N > 100) * 8 * log(N) / D)} # automatically choosing number of bins for raw X
    if(NbinsX > 0) # if a pariticular number of bins is specified 
    {
      if(D == 1){DataBinned <- binning(x = as.vector(X), y = as.vector(Y), nbins = NbinsX)}
      if(D == 2){DataBinned <- binning(x = X, y = Y, nbins = NbinsX)}			
    }
    if(NbinsX == 0) # no binning
    {
      DataBinned <- list()
      DataBinned$x <- X # design matrix is the raw data
      DataBinned$means <- Y # response is the raw Y
      DataBinned$devs <- rep(0, N) # no deviation since no binning taking place
      DataBinned$x.freq <- rep(1, N) # all weights set to 1
    }
    
    # compute bandwidth
    
    BWMethod <- toupper(BWMethod)
    if(BWMethod == "DF" & is.na(DF)){DF <- switch(D, 6, 12)}
    LossFunName <- paste(BWMethod, "LocSmoothing", sep = "")
    
    if(D == 1)
    {
      Start <- (sd(X) / 16)	
      End <- (sd(X) * 2)
      OptimList <- optimise(f = get(LossFunName), Data = DataBinned, Structure.2D = Structure.2D, LOPO = LOPO, LocLinQua = LocLinQua, LLQParams = LLQParams, DF = DF, interval = log(c(Start, End)))
      BW <- exp(OptimList$minimum)      
    }else
    {
      Start <- switch(Structure.2D, common = mean(sd(X) / 16), scaled = 0.5 / 8, separate = sd(X) / 16)
      End <- switch(Structure.2D, common = mean(sd(X) * 2), scaled = 0.5 * 4, separate = sd(X) * 2)
      
      if(Structure.2D != "separate")
      {
        OptimList <- optimise(f = get(LossFunName), Data = DataBinned, Structure.2D = Structure.2D, LOPO = LOPO, LocLinQua = LocLinQua, LLQParams = LLQParams, DF = DF, interval = log(c(Start, End)))
      }else
      {
        OptimList <- optim(par = log(Start*8), control = list(reltol = 1e-06), fn = get(LossFunName), Data = DataBinned, Structure.2D = Structure.2D, LOPO = LOPO, LocLinQua = LocLinQua, LLQParams = LLQParams, DF = DF)
      }
      BW <- switch(Structure.2D, common = exp(rep(OptimList$minimum, 2)), 
                                 scaled = sd(X) * exp(rep(OptimList$minimum, 2)),
                                 separate = exp(OptimList$par))
    }
    
    # return bandwidth
    
    return(BW)
  }
  
  
  
















