  
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  ####################################################################################
  ####    Function of Local Linear-Quadratic Combined 2 Dimensional Regression    ####
  ####################################################################################
  ####              version 1.0 Created by Chong Liu on 2012/11/04                ####
  ####################################################################################
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#                                              
  # Input
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.X:          Design matrix of size N by D
  # 2.Y:          Response variable matrix of size N by 1
  # 3.X0:         Evaluation points matrix of size N0 by D
  # 4.H:          Bandwidth used in kernel
  # 5.NbinsX:     number of bins used for design matrix X
  #               NbinsX = -1, automatically determine the number of bins used according to SM package
  #               NbinsX = 0, no binning
  #               NbinsX is positive integer, number of bins used
  # 6.NbinsX0:    number of bins used for evaluation points matrix X0
  #               NbinsX0 = -1, automatically determine the number of bins used according to SM package
  #               NbinsX0 = 0, no binning
  #               NbinsX0 is positive integer, number of bins used  
  # 7.Method:     Methods used for local linear-quadratic smoothing
  #               "poly" uses (1,X1-X0,X2-X0,(X2-X0)^2 and MA750 type of polynomial expressions to solve for the local fitting
  #               "matrix" uses (1,X1,X2,X2^2) and compact matrix expressions to solve for the local fitting
  # 8.InvSym:     A facas object in Ryacas package of matrix inverse used for corresponding method indicated by Method. Reason we need this argument is symbolic 
  #               inverse of 4 by 4 matrix takes a while. In simulations, this symbolic inverse is the same for each simulated data but evaluated at different points. 
  #               It is time consuming to do this inverse each time. Instead, it can be provided from outside and we only evaluate the inverse object at different points
  #               using this function in simulations. 
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Output
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.M:          Regression function evaluated at X0 or binned X0
  # 2.X.bin:      Binned X (may be the same as input X)
  # 3.Y.bin:      Binned Y (may be the same as input Y)
  # 4.X0.bin:     Binned X0 (may be the same as input X0)
  # 5.Weights:    Weights used in smoothing
  # 6.Devs:       If binning is used, the sum of square deviations of response Y from the average within each bin of X
  # 7.NbinsX:     Final number of bins used for X
  # 8.NbinsX0:    Final number of bins used for X0
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#


	LocLinQuaReg2D <- function(X, Y, X0, H, Weights = NULL, NbinsX = -1, NbinsX0 = 0, Method = "matrix", InvSym = NULL)
	{

		# load packages and functions

			#source("KerEval.r")
			#source("U.RepMat.r")
      #source("U.InvMat.r")
      
    # check data consistency
    
      if(!is.matrix(X) | !is.matrix(Y) | !is.matrix(X0)){stop("input must be in matrix format")}  
    
		# summarize raw data

			X.raw <- X # construct the raw X data
			Y.raw <- Y # construct the raw Y data
			N.raw <- dim(X.raw)[1] # number of data points in raw data
	
		# binning design matrix X

			if(NbinsX == -1){NbinsX <- round((N.raw > 100) * 8 * log(N.raw) / 2)} # automatically choosing number of bins for raw X
			if(NbinsX > 0) # if a pariticular number of bins is specified 
			{
				tempdata <- binning(x = X.raw, y = Y.raw, nbins = NbinsX)
				X <- as.matrix(tempdata$x) # X variable after binning
				Y <- as.matrix(tempdata$means) # response variable after binning, which is the average of all raw Y values in each bin
				Devs <- tempdata$devs # the sum of square deviations of response Y to its mean within each bin
				if(!is.null(Weights)){warning("Weights are masked by binning effect")} # user provided weights are masked by the binning Weights
				Weights <- tempdata$x.freq
			}
			if(NbinsX == 0) # no binning
      {
        X <- X.raw # design matrix is the raw data
        Y <- Y.raw # response is the raw Y
				Devs <- rep(0, N.raw) # no deviation since no binning taking place
			}
	
		# binning evaluation points X0

			X0.raw <- X0 # construct the raw X0
			N0.raw <- length(X0) # number of data points in raw X0
			if(NbinsX0 == -1){NbinsX0 <- round((N0.raw > 100) * 8 * log(N0.raw) / 2)} # automatically choose number of bins for raw X0
			if(NbinsX0 > 0) # if a pariticular number of bin is specified for X0
			{
				tempdata <- binning(x = X0.raw, nbins = NbinsX0)
				X0 <- as.matrix(tempdata$x)
			}
			if(NbinsX0 == 0){X0 <- X0.raw} # no binning

		# do the regression
       
      Method <- tolower(Method)      
      N <- dim(X)[1] # number of design points after binning
      if(is.null(Weights)){Weights <- rep(1, N)} # assign weights if not specified by user    
      N0 <- dim(X0)[1] # number of evaluation points after binning				
      K <- KerEval(X = X, X0 = X0, H = H) # kernel evaluation matrix of size N0 by N
      K <- t(Weights * t(K)) # multiply the kernel evaluation by extra weights, with size N0 by N
      K[abs(K) < 1e-20] <- 0
      
      if(Method == "matrix")
      {
  			XX <- cbind(rep(1,N), X, X[,2]^2) # construct fitting design matrix by combining X variable with intercept vector and the second order term X2^2 to form (1,X1,X2,X2^2)
  			XX0 <- cbind(rep(1,N0), X0, X0[,2]^2) # construct evaluation design matrix by combining X0 with intercept vector and the second order term X2^2 to form (1,X1,X2,X2^2)
  			temp1 <- XX %x% t(rep(1,4))
  			temp2 <- t(rep(1,4)) %x% XX
  			XXtilde <- temp1 * temp2 # compute the X'X part in (X'KX) as separate columns to get a matrix of size N by 16
  			PreInv <- K %*% XXtilde # insert the kernel part to get a matrix of size N0 by 16, with each column being the specific positin in (X'KX) evaluated at all X0        
        MatSym <- matrix(paste("A", 1 : 16, sep = ""), 4, 4, byrow = TRUE)
        INV <- InvMat(MatSym = MatSym, InvSym = InvSym, Entry = PreInv, Output = c("Inverse"), InverseFold = FALSE)$Inverse
				X1tilde <- t(as.vector(XX[,1]) * t(K)) 
				X2tilde <- t(as.vector(XX[,2]) * t(K)) 
  			X3tilde <- t(as.vector(XX[,3]) * t(K))			
  			X4tilde <- t(as.vector(XX[,4]) * t(K))			
  			XX1tilde <- INV[,1] * X1tilde + INV[,2] * X2tilde + INV[,3] * X3tilde + INV[,4] * X4tilde
				XX2tilde <- INV[,5] * X1tilde + INV[,6] * X2tilde + INV[,7] * X3tilde + INV[,8] * X4tilde
  			XX3tilde <- INV[,9] * X1tilde + INV[,10] * X2tilde + INV[,11] * X3tilde	+ INV[,12] * X4tilde				
  			XX4tilde <- INV[,13] * X1tilde + INV[,14] * X2tilde + INV[,15] * X3tilde + INV[,16] * X4tilde					
  			HatMat <- XX0[,1] * XX1tilde + XX0[,2] * XX2tilde + XX0[,3] * XX3tilde + XX0[,4] * XX4tilde
			}		
      
      if(Method == "poly")
      {
        MatSym <- matrix(c("A1","A2","A3","A4","A2","A5","A6","A7","A3","A6","A4","A8","A4","A7","A8","A9"), 4, 4, byrow = TRUE)
        temp1 <- RepMat(t(X[,1]), N0, 1) - RepMat(as.matrix(X0[,1]), 1, N) # matrix of size N0 by N
        temp2 <- RepMat(t(X[,2]), N0, 1) - RepMat(as.matrix(X0[,2]), 1, N) # matrix of size N0 by N
        AEval <- matrix(NA, N0, 9)
        AEval[,1] <- rowSums(K) # vector of length N0
        AEval[,2] <- rowSums(K * temp1) # vector of length N0
        AEval[,3] <- rowSums(K * temp2) # vector of length N0
        AEval[,4] <- rowSums(K * temp2 * temp2) # vector of length N0
        AEval[,5] <- rowSums(K * temp1 * temp1) # vector of length N0
        AEval[,6] <- rowSums(K * temp1 * temp2) # vector of length N0
        AEval[,7] <- rowSums(K * temp1 * temp2 * temp2) # vector of length N0
        AEval[,8] <- rowSums(K * temp2 * temp2 * temp2) # vector of length N0
        AEval[,9] <- rowSums(K * temp2 * temp2 * temp2 * temp2) # vector of length N0 
        INV <- InvMat(MatSym = MatSym, InvSym = InvSym, Entry = AEval, Output = c("Inverse"), InverseFold = FALSE)$Inverse # not the same INV as in method "matrix"
        B1 <- K # matrix of size N0 * N
        B2 <- K * temp1 # matrix of size N0 * N
        B3 <- K * temp2 # matrix of size N0 * N
        B4 <- K * temp2 * temp2 # matrix of size N0 * N
        HatMat <- INV[,1] * B1 + INV[,2] * B2 + INV[,3] * B3 + INV[,4] * B4
      }
			
			M <- HatMat %*% Y

    # return regression results
        
		  return(list(M = M, H = HatMat, X.bin = X, Y.bin = Y, X0.bin = X0, Weights = Weights, Devs = Devs, NbinsX = NbinsX, NbinsX0 = NbinsX0))
	}