  
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  #############################################################
  ####         Function of Local Linear Regression         ####
  #################################################3###########
  ####   version 1.0 Created by Chong Liu on 2012/10/31    ####
  #############################################################
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#                                              
  # Input
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.X:          Design matrix of size N by D
  # 2.Y:          Response variable matrix of size N by 1
  # 3.X0:         Evaluation points matrix of size N0 by D
  # 4.H:          Bandwidth used in kernel
  # 5.NbinsX:     number of bins used for design matrix X
  #               NbinsX = -1, automatically determine the number of bins used according to SM package
  #               NbinsX = 0, no binning
  #               NbinsX is positive integer, number of bins used
  # 6.NbinsX0:    number of bins used for evaluation points matrix X0
  #               NbinsX0 = -1, automatically determine the number of bins used according to SM package
  #               NbinsX0 = 0, no binning
  #               NbinsX0 is positive integer, number of bins used  
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Output
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.M:          Regression function evaluated at X0 or binned X0
  # 2.X.bin:      Binned X (may be the same as input X)
  # 3.Y.bin:      Binned Y (may be the same as input Y)
  # 4.X0.bin:     Binned X0 (may be the same as input X0)
  # 5.Weights:    Weights used in smoothing
  # 6.Devs:       If binning is used, the sum of square deviations of response Y from the average within each bin of X
  # 7.NbinsX:     Final number of bins used for X
  # 8.NbinsX0:    Final number of bins used for X0
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#


	LocLinReg <- function(X, Y, X0, H, Weights = NULL, NbinsX = -1, NbinsX0 = 0)
	{

		# load packages and functions

			#source("U.KerEval.r")
			#source("U.RepMat.r")
		
    # check data consistency
    
      if(!is.matrix(X) | !is.matrix(Y) | !is.matrix(X0)){stop("input must be in matrix format")}  
    
		# summarize raw data

			X.raw <- X # construct the raw X data
			Y.raw <- Y # construct the raw Y data
			N.raw <- dim(X.raw)[1] # number of data points in raw data
			D <- dim(X.raw)[2] # dimension of regression

		# binning design matrix X

			if(NbinsX == -1){NbinsX <- round((N.raw > 100) * 8 * log(N.raw) / D)} # automatically choosing number of bins for raw X
			if(NbinsX > 0) # if a pariticular number of bins is specified 
			{
				if(D == 1){tempdata <- binning(x = as.vector(X.raw), y = as.vector(Y.raw), nbins = NbinsX)}
				if(D == 2){tempdata <- binning(x = X.raw, y = Y.raw, nbins = NbinsX)}
				X <- as.matrix(tempdata$x) # design matrix after binning
				Y <- as.matrix(tempdata$means) # response variable after binning, which is the average of all raw Y values in each bin
				Devs <- tempdata$devs # the sum of square deviations of response Y to its mean within each bin
				if(!is.null(Weights)){warning("Weights are masked by binning effect")} # user provided weights are masked by the binning Weights
				Weights <- tempdata$x.freq				
			}
			if(NbinsX == 0) # no binning
      {
        X <- X.raw # design matrix is the raw data
        Y <- Y.raw # response is the raw Y
				Devs <- rep(0, N.raw) # no deviation since no binning taking place
			}

		# binning evaluation points X0

			X0.raw <- X0 # construct the raw X0
			N0.raw <- length(X0) # number of data points in raw X0
			if(NbinsX0 == -1){NbinsX0 <- round((N0.raw > 100) * 8 * log(N0.raw) / D)} # automatically choose number of bins for raw X0
			if(NbinsX0 > 0) # if a pariticular number of bin is specified for X0
			{
				if(D == 1){tempdata <- binning(x = as.vector(X0.raw), nbins = NbinsX0)}
				if(D == 2){tempdata <- binning(x = X0.raw, nbins = NbinsX0)}
				X0 <- as.matrix(tempdata$x)
			}
			if(NbinsX0 == 0){X0 <- X0.raw} # no binning

		# do the regression
	
			N <- dim(X)[1] # number of design points after binning
			N0 <- dim(X0)[1] # number of evaluation points after binning
      if(is.null(Weights)){Weights <- rep(1, N)} # assign weights if not specified by user or forced to be by binning process
			K <- KerEval(X = X, X0 = X0, H = H) # kernel evaluation matrix of size N0 by N
			K <- t(Weights * t(K)) # multiply the kernel evaluation by extra weights			
			XX <- cbind(rep(1,N), X) # combine design points with intercept vector
			XX0 <- cbind(rep(1,N0), X0) # combine evaluation points with intercept vector
        
			temp1 <- XX %x% t(rep(1, (D+1)))
			temp2 <- t(rep(1, (D+1))) %x% XX
			XXtilde <- temp1 * temp2
			preINV <- K %*% XXtilde
			INV <- array(NA, dim = c(N0,(D+1)^2))
			if(D == 1)
			{
				Det <- preINV[,1]*preINV[,4] - preINV[,2]*preINV[,3]
				INV[,1] <- preINV[,4]
				INV[,2] <- -preINV[,2]
				INV[,3] <- INV[,2]
				INV[,4] <- preINV[,1]			
				X1tilde <- t(as.vector(XX[,1]) * t(K)) 
				X2tilde <- t(as.vector(XX[,2]) * t(K))	
				XX1tilde <- INV[,1] * X1tilde + INV[,2] * X2tilde
				XX2tilde <- INV[,3] * X1tilde + INV[,4] * X2tilde					
				HatMat <- XX0[,1] * XX1tilde + XX0[,2] * XX2tilde
				HatMat <- HatMat / Det			
			}
      if(D == 2)
			{
				Det <- preINV[, 1]*preINV[,5]*preINV[,9] + preINV[,3]*preINV[,4]*preINV[,8] + preINV[,2]*preINV[,6]*preINV[,7] - preINV[,3]*preINV[,5]*preINV[,7] - preINV[,1]*preINV[,6]*preINV[,8] - preINV[,2]*preINV[,4]*preINV[,9]
				INV[,1] <- (preINV[,5]*preINV[,9] - preINV[,6]*preINV[,8]) 
				INV[,2] <- (preINV[,3]*preINV[,8] - preINV[,2]*preINV[,9]) 
				INV[,3] <- (preINV[,2]*preINV[,6] - preINV[,3]*preINV[,5]) 
				INV[,5] <- (preINV[,1]*preINV[,9] - preINV[,3]*preINV[,7]) 
				INV[,6] <- (preINV[,3]*preINV[,4] - preINV[,1]*preINV[,6]) 
				INV[,9] <- (preINV[,1]*preINV[,5] - preINV[,2]*preINV[,4])
				INV[,4] <- INV[,2]
				INV[,7] <- INV[,3]
				INV[,8] <- INV[,6]	
				X1tilde <- t(as.vector(XX[,1]) * t(K)) 
				X2tilde <- t(as.vector(XX[,2]) * t(K)) 
				X3tilde <- t(as.vector(XX[,3]) * t(K))			
				XX1tilde <- INV[,1] * X1tilde + INV[,2] * X2tilde + INV[,3] * X3tilde
				XX2tilde <- INV[,4] * X1tilde + INV[,5] * X2tilde + INV[,6] * X3tilde
				XX3tilde <- INV[,7] * X1tilde + INV[,8] * X2tilde + INV[,9] * X3tilde					
				HatMat <- XX0[,1] * XX1tilde + XX0[,2] * XX2tilde + XX0[,3] * XX3tilde
				HatMat <- HatMat / Det
			}			
			
			M <- HatMat %*% Y

    # return regression results
        
		  return(list(M = M, H = HatMat, X.bin = X, Y.bin = Y, X0.bin = X0, Weights = Weights, Devs = Devs, NbinsX = NbinsX, NbinsX0 = NbinsX0))
	}