  
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  ##################################################################################
  ####    Function to Generate Neighborhood for Nonparametric XCov Smoothing    ####
  ##################################################################################
  ####             version 2.0 Created by Chong Liu on 2012/11/15               ####
  ##################################################################################
  #### change from version 1.0                                                  ####
  #### 1.add CurveID                                                            ####
  ##################################################################################
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#                                              
  # Input:
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.Coordinate:       The usual cartesian coordinate of all curve locations
  # 2.SpaSepType:       Type of spatial sepration used in the SpaSep.
  #                     "cartesian" is the usual cartesian distance separation vector. The second dimension in SpaSep correspond to x and y axis;
  #                     "polar" is the polar coordinates. The second dimension in SpaSep correspond to distance and angle.
  # 3.SpaSep:           A matrix of size NTarget by 2 by 2. The first dimension is the number of neighborhoods in consideration (number of XCov need to be fitted). The second dimension 
  #                     is the end points of the interval used to determine neighborhoods. The third dimension is for two separation variables a and b.
  # 4.CurveID:          IDs of curves involved
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Output
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.NeighborPairs:    A list of length NTarget in which each cell stores the pairs of locations whose spatial variables fall into the target interval specified in the corresponding
  #                     entries in SpaSep.
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Details
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # All points that are polar symmetric is not distinguishable in this neighborhood structure, so spatial separation vectors that are polar symmetric are treated as the same
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  
  
  NeighborGen <- function(Coordinate, SpaSepType, SpaSep, CurveID = NULL)
  {
    # source functions
       
      #source("SCorGen.r")
      #source("AngleGen.r")
      #source("DistGen.r") 
      
    # Construct spatial separation coordinate matrices according to the SpaSepType
      
      SpaSepType <- tolower(SpaSepType)
      
      if(SpaSepType == "cartesian")
      {
        coordinate.x <- as.matrix(Coordinate[,1]) # x coordinate
        coordinate.y <- as.matrix(Coordinate[,2]) # y coordinate
        N <- length(coordinate.x) # number of curves
        SpaSep.a <-  RepMat(t(coordinate.x),N,1) - RepMat(coordinate.x,1,N) # spatial separation on x axis
        SpaSep.b <- RepMat(t(coordinate.y),N,1) - RepMat(coordinate.y,1,N) # spatial separation on y axis
        SpaSep.b[SpaSep.a < 0] <- -SpaSep.b[SpaSep.a < 0] # transform the negative x separation vector to its polar symmetric point in y direction      
        SpaSep.a[SpaSep.a < 0] <- -SpaSep.a[SpaSep.a < 0] # transform the negative x separation vector to its polar symmetric point in x direction
      }
      
      if(SpaSepType == "polar")
      {
        SpaSep.a <- DistGen(Coordinate = Coordinate, DistType = "euclidean") # separation in distance
        SpaSep.b <- AngleGen(Coordinate = Coordinate, Type = 1) # separation in angle
        SpaSep.b[SpaSep.b == -90] <- 90 # make the matrix symmetric, transform the -90 separation to its polar symmetric point
      }             

    # Construct Neighborhoods
      
      if(is.null(CurveID)){CurveID <- 1 : dim(Coordinate)[1]}
      NTarget <- dim(SpaSep)[1] # number of neighborhood targets
      NeighborPairs <- list() 
      
      for(t in 1 : NTarget)
      {
        thisTarget.a <- SpaSep[t,,1] # extract the current neighborhood target for separation variable a
        thisTarget.b <- SpaSep[t,,2] # extract the current neighborhood target for separation variable b
        if(thisTarget.a[1] == thisTarget.a[2]) # logical matrix indicating which pair of locations satisfying the current neghborhood target interval for spatial separation variable a
        {
          neighbor.a <- (SpaSep.a == thisTarget.a[1])
        }else
        {
          neighbor.a <- (SpaSep.a > thisTarget.a[1] & SpaSep.a <= thisTarget.a[2])
        }  
        if(thisTarget.b[1] == thisTarget.b[2]) # logical matrix indicating which pair of locations satisfying the current neghborhood target interval for spatial separation variable b
        {
          neighbor.b <- (SpaSep.b == thisTarget.b[1])
        }else
        {
          neighbor.b <- (SpaSep.b > thisTarget.b[1] & SpaSep.b <= thisTarget.b[2])
        }         
        neighbor <- neighbor.a & neighbor.b # take neighborhood pairs which satisfy both targets in variable a and b, still a logical matrix
        if(all(!neighbor))
        {
          stop("No available neighbors") # no TRUE in the matrix
        }
        NeighborPairs[[t]] <- t(apply(which(t(neighbor), arr.ind = TRUE), MARGIN = 1, FUN = rev)) # all pairs of locations which satisfy target in both a and b
        NeighborPairs[[t]] <- unique(rbind(NeighborPairs[[t]],cbind(NeighborPairs[[t]][,2],NeighborPairs[[t]][,1]))) 
        # to make sure every pair in reverse order is included. A double check of treating polar symmetric points as the same
        NeighborPairs[[t]] <- NeighborPairs[[t]][order(NeighborPairs[[t]][,1]),]
        NeighborPairs[[t]] <- cbind(CurveID[NeighborPairs[[t]][,1]], CurveID[NeighborPairs[[t]][,2]]) # embed the index into Curve IDs
      }
    
    # return neighborhoods
      
      return(NeighborPairs)
  } 
         
  
  