
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  ##########################################################################
  ####    Function to Estimate Correlation with Projected Eigenvalues   ####
  ##########################################################################
  ####          version 1.0 Created by Chong Liu on 2012/11/14          ####
  ##########################################################################
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#                                              
  # Input:
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.CovFitMat:          A matrix of size NGrid by NGrid, the covariance surface
  # 2.XCovFitMat:         An array of size NTarget by NGrid by NGrid, the cross-covariance surface
  # 3.K:                  Number of eigenvalues and eigenfunctions retained
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Output
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.RhoEst.Proj:        A matrix of NTarget by K, the ratio between XLambdaEst.Proj and LambdaEst.Proj
  # 2.LambdaEst.Proj:     A vector of length K, the estimated eigenvalues of CovFitMat by projection method
  # 3.XLambdaEst.Proj:    A matrix of NTarget by K, estimated eigenvalues of XCovFitMat by projection method
  # 4.EigenFunEst.Proj:   A matrix of NGrid by K, estimated eigenvectors of the combined surfaces
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  
  RhoEstProj <- function(CovFitMat, XCovFitMat, K) 
  {
    # compute projected eigenvalues and correlation estimates
    
      if(length(dim(XCovFitMat)) == 2){XCovFitMat <- array(XCovFitMat, dim = c(1, dim(XCovFitMat)))}
      NTarget <- dim(XCovFitMat)[1] # number of targets in use
      CombineMat <- CovFitMat + apply(XCovFitMat, MARGIN = c(2,3), FUN = sum)
      eigs <- eigen(CombineMat)
      eigs$values[eigs$values < 1e-8] <- 0 # make negative and small positive eigenvalues to be zeros
      EigenFunProj <- eigs$vectors[, 1:K] # store the first K eigenfunctions common for all surfaces
      XLambdaProj <- matrix(NA, NTarget, K) # projected eigenvalues
      LambdaProj <- diag(t(EigenFunProj) %*% CovFitMat %*% EigenFunProj)
      for(t in 1 : NTarget)
      {
        XLambdaProj[t,] <- diag(t(EigenFunProj) %*% XCovFitMat[t,,] %*% EigenFunProj)
      }
      RhoEst.Proj <- XLambdaProj / RepMat(t(LambdaProj), NTarget, 1) # ratio of projected eigenvalues
    
    # return results
      
      return(list(RhoEst.Proj = RhoEst.Proj, LambdaEst.Proj = LambdaProj, XLambdaEst.Proj = XLambdaProj, EigenFunEst.Proj = EigenFunProj))
    
  }