
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  ##############################################################
  ####    Function to Generate Spatially Correlated Data    ####
  ##############################################################
  ####    version 3.0 Created by Chong Liu on 2012/11/18    ####
  ##############################################################
  #### changes of version 2.0 from version 1.0:             ####
  #### 1. Add generating random times on regular grids      #### 
  ##############################################################
  #### changes of version 3.0 from version 2.0:             ####
  #### 1. Change Alpha and Delta into lists                 #### 
  #### 2. Add output of fPC scores                          ####
  ##############################################################
  #### changes from version 3.0:                            ####
  #### 1. Make Alpha and Delta depend on CorType            #### 
  #### 2. Remove full model, Cortype                        ####
  #### 3. Remove the output CorLengend                      ####
  ##############################################################
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#                                              
  # Input:
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
	# 1.DistType:       Possible choices for distance type: "euclidean","manhattan","mahalanobis".
  # 2.AlphaList:      Angle (in degrees) for axes rotation in "mahalanobis" distance, a list of length A.
  #                   Each element of the list is a vector of length K.
  # 3.DeltaList:      Rescaling amount of axes in "mahalanobis" distance, a list of length A.
  #                   Each element of the list is a vector of length K.
	# 4.Coordinate:     Spatial coordinates of the curves.
	# 5.KappaList:      Kappa in the matern model. Suppose there are A cases and K eigenfunctions. Then KappaList is a list of length A. 
	#                   Each element of the list is a vector of length K.
	# 6.PhiList:        Phi in the matern model. Suppose there are A cases and K eigenfunctions. Then PhiList is a list of length A. 
	#                   Each element of the list is a vector of length K.
	# 7.TimeRange:      The time span of the curve.
	# 8.MPerCurve:      Number of points per curve.
	# 9.SigmaVec:       Standard deviation of noise.
	# 10.B:             Number of replicated data sets.
	# 11.MeanFD:        Mean curve in functional object format.
	# 12.EigenFun:      Eigenfunctions in functional object format.
	# 13.Lambda:        Eigenvalues of length K.
  # 14.TimeRegular:   A logical value indicating whether time points are taken on regular grids
  # 15.Grids:         If TimeRegular is true, the regular grids used to generate random times, with smallest grid point equal to left end of TimeRange and largest
  #                   grid point equal to the right end of TimeRange.
  # 16.Verbose:       Logical value indicating whether to show data generating process
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Output
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.DistMat:        A list of length A, which stores the distance matrix. Each element of the list is an array of size K by NCurve by NCurve.
	# 2.CorMat:         A list of correlation matrix of length A(the number of cases). Each element of the list is an array of size K by NCurve by NCurve.
	# 3.Curves:         A list of size B by A, the noiseless curves in the format of functional object
	# 4.Time:           A list of size B by A by S by M. Each cell of the list is also a list of length Ncurve. Each subcell stores the random evaluation time points on each individual 
  #                   curve within the TimeRange.
	# 5.Signal:         A list of size B by A by S by M. Each cell of the list is also a list of length Ncurve. Each subcell stores the evaluation of each individual curve at corresponding
  #                   time points.
	# 6.Observation:    A list of size B by A by S by M. Each cell of the list is also a list of length Ncurve. Each subcell stores the evaluation of each individual curve at corresponding
  #                   time points plus noise.
  # 7.Scores:         A matrix of size NCurve by K by B by A, the fPC scores of all curves and all eigenfunctions
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#

  SCorDatGen <- function(DistType, AlphaList, DeltaList, Coordinate, KappaList, PhiList, TimeRange, MPerCurve, SigmaVec, B, MeanFD, EigenFun, 
                         Lambda, TimeRegular, Grids = NULL, Verbose = FALSE)
	{

		#source("SCorFunGen.r")
		#source("DistGen.r")
    library(geoR)
    
    # check consistency
    
      if(TimeRegular){if(is.null(Grids) | min(Grids) != TimeRange[1] | max(Grids) != TimeRange[2]){stop("Grids are not correct")}}

		# generate correlation matrix and noiseless curves

			A <- length(PhiList) # number of different parameter setting
			CorMat <- array(list(), dim = c(A)) # a list of length A
      DistMat <- array(list(), dim = c(A)) # a list of length A
			Curves <- array(list(), dim = c(B,A))
      NCurve <- dim(Coordinate)[1] # number of curves in each single simulated data set
      K <- length(Lambda) # number of eigenfunctions
      Scores <- array(NA, dim = c(NCurve,K,B,A)) # fPC scores container
			for(a in 1 : A)
			{		
        thisAlpha <- AlphaList[[a]] # a vector of length K
        thisDelta <- DeltaList[[a]] # a vector of length K
			  thisPhi <- PhiList[[a]] # a vector of length K
			  thisKappa <- KappaList[[a]] # a vector of length K
        DistMat[[a]] <- array(NA, dim = c(K,NCurve,NCurve))
        CorMat[[a]] <- array(NA, dim = c(K,NCurve,NCurve))
        if(length(thisAlpha) != K){stop("Parameters dimension is not correct")}
        for(k in 1 : K)
        {
          DistMat[[a]][k,,] <- DistGen(Coordinate = Coordinate, DistType = DistType, Alpha = thisAlpha[k], Delta = thisDelta[k]) # the distance matrix of current Alpha and Delta
          CorMat[[a]][k,,] <- matern(u = DistMat[[a]][k,,], phi = thisPhi[k], kappa = thisKappa[k]) # the distance matrix of current Phi and Kappa
        }
        
        CurveList <- SCorFunGen(CorMat = CorMat[[a]], MeanFD = MeanFD, EigenFun = EigenFun, Lambda = Lambda, B = B) # generate noiseless curves
				Curves[,a] <- CurveList$Curves
        Scores[,,,a] <- CurveList$Scores
			}

		# generate discrete observations

			M <- length(MPerCurve)
			S <- length(SigmaVec)
			Signal <- array(list(), dim = c(B, A, S, M))
			Observation <- array(list(), dim = c(B, A, S, M))
			Time <- array(list(), dim = c(B, A, S, M))	
			for(b in 1 : B)
			{
				for(a in 1 : A)
				{
					for(s in 1 : S)
					{
						for(m in 1 : M)
						{
							Time[[b,a,s,m]] <- list()
							Signal[[b,a,s,m]] <- list()
							Observation[[b,a,s,m]] <- list()
							thisM <- MPerCurve[m]
							for(i in 1 : NCurve)
							{
								if(!TimeRegular)
                {
                  Time[[b,a,s,m]][[i]] <- sort((runif(thisM) * (TimeRange[2] - TimeRange[1])) + TimeRange[1])
                }else
								{
                  NGrid <- length(Grids)
								  Time[[b,a,s,m]][[i]] <- Grids[sort(sample(1 : NGrid, thisM))]
								}
								Signal[[b,a,s,m]][[i]] <- eval.fd(Time[[b,a,s,m]][[i]], Curves[[b, a]][i])
								Observation[[b,a,s,m]][[i]] <- Signal[[b,a,s,m]][[i]] + rnorm(thisM) * SigmaVec[s]
							}	
						}
					}
				}
        if(Verbose){print(paste("b =", b))}
			}		
	 			
		# return output

		  output <- list(DistMat = DistMat, CorMat = CorMat, Curves = Curves, Scores = Scores, Signal = Signal, Observation = Observation, Time = Time)	
		  return(output)
		
		}

