  
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  ###############################################################################################
  ####    Function to Estimate Anisotropy Matern Model Parameters Using Method of Moments    ####
  ###############################################################################################
  ####                   version 5.0 Created by Chong Liu on 2012/11/25                      ####
  ###############################################################################################
  #### change of version 2.0 from version 1.0                                                ####
  #### 1.add projection method of computing eigenvalues and eigenfunctions                   ####
  #### 2.use AnisoMaternFit function to fit correlation model                                ####
  ###############################################################################################
  #### change of version 3.0 from version 2.0                                                ####
  #### 1.use SmoothCore                                                                      ####
  #### 2.use RhoEstProj                                                                      ####
  ###############################################################################################
  #### change of version 4.0 from version 3.0                                                ####
  #### 1.add CV on K and Targets                                                             ####
  ###############################################################################################
  #### change from version 4.0                                                               ####
  #### 1.fit model with optimal K and NTarget for each task                                  ####
  #### 2.add fitting on all possible targets                                                 ####
  #### 3.use new Sigma fitting function in SmoothCov                                         ####
  ###############################################################################################
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#                                              
  # Input:
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.Y:                    Observations of underlying random function.
  #                         If a matrix, Y is observed at regular grids of time points, however, maybe with missingness. 
  #                         Rows correspond to locations and columns correspond to time points.
  #                         If a List, Y is observed at irregular time points. The list is of length Ncurve in which each cell stores the observations on each individual curve
  # 2.Time:                 Time points corresponding to Y, in the same format as Y
  # 3.TimeRange:            The time range of Time, used to normalize Time onto [0,1] interval
  # 4.TimeRegular:          A logical variable indicating whether observation times shoudl be coerced to the nearest regular time points according to EvalTime.Regular
  # 5.EvalTime.Regular:     Valuation times used to approximate real time
  # 6.Rounding:             Way of approximating real time using EvalTime.Regular
  #                         "up" approximates each real time point to the nearest EvalTime from above
  #                         "down" approximates each real time point to the nearest EvalTime from below
  #                         "cloest" approximates each real time point to the nearest EvalTime from either side
  # 7.EvalTime.Mu:          A vector of length NGrid which is the time points used to evaluate fitted mean
  # 8.KAll:                 A vector, each entry of which is the number of eigenfunctions retained in results.
  # 9.EvalTime.Diag:        A vector of time points used to evaluate diagonal component of the covariance surface
  # 10.LLQParams:           A list of parameters for local linear quadratic smoothing, including
  #                         Method:    Methods used for local linear-quadratic smoothing
  #                                    "poly" uses (1,X1-X0,X2-X0,(X2-X0)^2 and MA750 type of polynomial expressions to solve for the local fitting
  #                                    "matrix" uses (1,X1,X2,X2^2) and compact matrix expressions to solve for the local fitting
  #                         InvSym:    A facas object in Ryacas package of matrix inverse used for corresponding method indicated by Method. 
  #                                    Reason we need this argument is symbolic inverse of 4 by 4 matrix takes a while. 
  #                                    In simulations, this symbolic inverse is the same for each simulated data but evaluated at different points. 
  #                                    It is time consuming to do this inverse each time. 
  #                                    Instead, it can be provided from outside and we only evaluate the inverse object at different points using this function in simulations.
  # 11.Coordinate:          Coordinates of the locations
  # 12.Rotate:              Whether to rotate the axis by 45 degree in smoothing covariance surface
  # 13.TargetType:          Type of neighborhood pair
  #                         "cartesian" is the usual cartesian distance separation vector. The second dimension in TargetAll correspond to x and y axis;
  #                         "polar" is the polar coordinates. The second dimension in TargetAll correspond to distance and angle.
  # 14.TargetAll:           An array of size NTarget by 2 by 2. 
  #                         The first dimension is the number of neighborhood targets to be considered (number of XCov need to be fitted). 
  #                         The second dimension is the end points of the interval used to determine neighborhoods. 
  #                         The third dimension is for two separation variables a and b.
  # 15.TargetIdxList:       A list of vectors. Each vector contains indices of neighborhood pairs to be used in model estimation. 
  #                         All vectors are explored in model selection and the optimal one is determined
  # 16.ModelSelection:      A logical value indicating whether to perform model selection based on CV procedure on number of eigenvalues retianed K, 
  #                         and which neighborhood pairs are used in model estimation
  # 17.TargetRep:           A matrix of size NTarget by 2, where each row is the representative locations for the corresponding neighborhood pair 
  # 18.DefaultPars:         A matrix of size NMethod by 4, where the 4 columns correspond to Alpha, Delta, Kappa, Phi. 
  #                         Non NA values are fixed in model fitting. 
  #                         NA values indicates the parameters to be fitted
  # 19.InitialPars:         A matrix of size NMethod by 4, where the 4 columns correspond to Alpha, Delta, Kappa, Phi. 
  #                         These are the starting values in parameter fitting optimization 
  # 20.ParsBound:           An array of size NMethod by 2 by 4. The first dimension is corresponding to optimization methods 
  #                         The second dimension is corresponding to the lower and upper bounds; The third dimension is corresponding to the 4 parameters
  # 21.Log:                 A matrix of size NMethod by 4, containing logical values of whether parameters are in log scale   
  # 22.Optim.Method:        A vector of length NTask, where each entry is the name of optimization method to be used for each task.
  #                         "Nelder-Mead", "BFGS", "L-BFGS-B", "nls", "Grid" . The first three methods are included in "optim" function 
  # 23.MaxIt:               A vector of length NTask, where each entry is the number of maximum number of iterations
  # 24.NGridsSearch:        A matrix of size NMethod by 4, containing the number of grid points used in "Grid" search optimization method
  # 25.Dim:                 Dimension of the curve spatial locations              
  # 26.ReConstructType:     Method of curve reconstruction
  #                         "seperable", assuming separable structure, use the fitted parameters of the first eigenvalue for all eigenvalues
  #                         "diag", use fitted parameters for each individual eigenvalue
  # 27.Sigma:               Back up sigma value if the fitted sigma is NaN
  # 28.CVMean:              Logical value indicating whether mean fitting is also considered in model selection
  # 29.D1NFold:             Number of folds for 1 dimensional locations
  # 30.D1Separation:        Bandwidth method
  # 31.SigmaMethod:         A vector of methods used for estimating Sigma
  #                         "PACE", the original PACE method 
  #                         "Modified PACE", the modified PACE method
  # 32.Verbose:             Logical value indicating whether each neighbor pair is finished
  # 33.BWMethod:            Bandwidth selection method
  # 34.DF:                  Degree of freedom used in DF method in choosing bandwidth
  # 35.ResModelK:           The way of combining PC scores which are assumed to follow the same spatial correlation structure in restricted model of hypothesis test
  #                         "no", then do not compute restricted model estimation
  #                         "opt", then the selected optimal number of K PC scores are combined to fit the restricted model
  #                         Any integer number indicating how many first PC scores are combined to fit the restricted model
  # 36.Reconstruction:      A logical value, indicating whether curve reconstruction will be performed
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Output
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#

    SCorEstMM <- function(Y, Time, SCorEstMMParams)
    {

      # load packages and functions
      
        # library(fda)
        # library(sm)
        # source("SmoothCore.r")
        # source("AnisoMaternFit.r")
      
      # Parameters
      
        TimeRange <- SCorEstMMParams$TimeRange
        TimeRegular <- SCorEstMMParams$TimeRegular 
        EvalTime.Regular <- SCorEstMMParams$EvalTime.Regular 
        Rounding <- SCorEstMMParams$Rounding 
        EvalTime.Mu <- SCorEstMMParams$EvalTime.Mu 
        KAll <- SCorEstMMParams$KAll 
        EvalTime.Diag <- SCorEstMMParams$EvalTime.Diag
        LLQParams <- SCorEstMMParams$LLQParams
        Coordinate <- SCorEstMMParams$Coordinate 
        Rotate <- SCorEstMMParams$Rotate
        TargetType <- SCorEstMMParams$TargetType
        TargetAll <- SCorEstMMParams$TargetAll
        TargetIdxList <- SCorEstMMParams$TargetIdxList
        ModelSelection <- SCorEstMMParams$ModelSelection 
        TargetRep <- SCorEstMMParams$TargetRep
        DefaultPars <- SCorEstMMParams$DefaultPars
        InitialPars <- SCorEstMMParams$InitialPars 
        ParsBound <- SCorEstMMParams$ParsBound 
        Log <- SCorEstMMParams$Log
        Optim.Method <- SCorEstMMParams$Optim.Method 
        MaxIt <- SCorEstMMParams$MaxIt 
        NGridsSearch <- SCorEstMMParams$NGridSearch 
        Dim <- SCorEstMMParams$Dim
        ReConstructType <- SCorEstMMParams$ReConstructType 
        Sigma <- SCorEstMMParams$Sigma 
        CVMean <- SCorEstMMParams$CVMean
        D1NFold <- SCorEstMMParams$D1NFold 
        D1Separation <- SCorEstMMParams$D1Separation 
        SigmaMethod <- SCorEstMMParams$SigmaMethod 
        Verbose <- SCorEstMMParams$Verbose 
        BWMethod <- SCorEstMMParams$BWMethod
        DF <- SCorEstMMParams$DF
        ResModelK <- SCorEstMMParams$ResModelK 
        Reconstruction <- SCorEstMMParams$Reconstruction
        
      # massage data into format for smoother
        
        DataList <- DataMassage(Y = Y, Time = Time, TimeRange = TimeRange, Format = "Stack", TimeRegular = TimeRegular, EvalTime = EvalTime.Regular, Rounding = Rounding) 
        #DataPerCurve <- DataMassage(Y = Y, Time = Time, TimeRange = TimeRange, Format = "PerCurve", TimeRegular = TimeRegular, EvalTime = EvalTime.Regular, Rounding = Rounding) 
        
        # original data is summarized in list where each cell stores the observations from an individual curve
        # this original data is massaged into stacked format
        if(TimeRegular)
        {
          Time <- as.matrix(DataList$TimeApprox) # the new time approximated by EvalTime.Regular if time on regular points is desired
          TimeIndex <- as.matrix(DataList$TimeApproxIdx) # the index of approximated time in EvalTime.Regular
        }else
        {
          Time <- as.matrix(DataList$Time) # if original time is desired, then the new time is the same as the original time
          TimeIndex <- NULL # if original time is desired, then no index is available.
        }
        Y <- as.matrix(DataList$Y) # the vector of Ys from all curves stacked together
        ID <- as.matrix(DataList$ID) # the vector of curve IDs from all curves stacked together
        DataMat <- cbind(ID, Y, Time) #  combine the stacked Y, ID and Time into a matrix
        NCurve <- length(unique(ID)) # number of curves
        NTask <- length(Optim.Method) # number of model fitting tasks
        
      # choose optimal K and optimal number of Targets for each task
        
        KOpt <- c() # container of optimal K, the number of eigenfunctions retained, for each task
        TarIdxOpt <- list() # container of optimal target neighborhood index, for each task
        
        if((length(KAll) == 1 & length(TargetIdxList) == 1) & ModelSelection) 
        {# if only one K and target index vector are provided and model selection is still desired, then use these parameters for model estimation. 
          print("Only one K and one Target provided with model selection")
          for(task in 1 : NTask)
          {            
            KOpt[task] <- KAll
            TarIdxOpt[[task]] <- TargetIdxList[[1]]           
          }
          ModelSelList <- NULL
        }else if(ModelSelection)
        {# more than one K or one target index vector is provided and model selection is provided, then cross-validation is performed to select optimal K and target index vector
          print("Cross-Validation in choosing K and Targets")
          CVFold <- CVTrainTestGen(Dim = Dim, Coordinate = Coordinate, D1NFold = D1NFold, D1Separation = D1Separation) # generate training and testing set of curves
          Train <- CVFold$Train # the list of training sets
          Test <- CVFold$Test # the list of testing sets 
          if(!CVMean)
          {
            HMu <- HSelect(X = as.matrix(Time), Y = as.matrix(Y), BWMethod = BWMethod, LOPO = 2, LocLinQua = FALSE, NbinsX = -1, DF = DF)      	
            FitMu <- LocLinReg(X = as.matrix(Time), Y = as.matrix(Y), X0 = as.matrix(Time), H = HMu, NbinsX = -1, NbinsX0 = 0)$M # evaluation at all time points in DataMat
          }else
          {
            FitMu <- NULL 
          }
          ModelSelList <- SCorEstMMCV(Train = Train, Test = Test, Coordinate = Coordinate, DataMat = DataMat, TargetAll = TargetAll, TargetType = TargetType, TargetIdxList = TargetIdxList,
                                      EvalTime.Mu = EvalTime.Mu, KAll = KAll, EvalTime.Diag = EvalTime.Diag, Rotate = Rotate, ReConstructType = ReConstructType, DefaultPars = DefaultPars,
                                      InitialPars = InitialPars, ParsBound = ParsBound, TargetRep = TargetRep, Optim.Method = Optim.Method, Log = Log, MaxIt = MaxIt, 
                                      NGridsSearch = NGridsSearch, Sigma = Sigma, TimeIndex = TimeIndex, FitMu = FitMu, CVMean = CVMean, SigmaMethod = SigmaMethod, LLQParams = LLQParams, 
                                      BWMethod = BWMethod, DF = DF, Verbose = Verbose) # perform model selection
          Optimal <- ModelSelList$Optimal
          # a matrix of size NTarget by 2, the 1st entry of which is the index of optimal target vector in TargetIdxList, the 2nd entry of which is the index of optimal K in KAll 
          for(task in 1 : NTask)
          {
            KOpt[task] <- KAll[Optimal[task,2]] # the optimal K, for each task
            TarIdxOpt[[task]] <- TargetIdxList[[Optimal[task,1]]] # the optimal target index vector, for each task
          }
        }else
        {
          print("No model selection. All neighborhoods in TargetAll and maximum K are used in model estimation")
          for(task in 1 : NTask)
          {
            KOpt[task] <- max(KAll) # select the largest K, for each task
            TarIdxOpt[[task]] <- 1:dim(TargetAll)[1] # select all targets in TargetAll, for each task
            ModelSelList <- NULL
          }
        } 
              
      # smooth mean, covariance surface and Xcovariance surfaces after model selection using union of Targets and K across all optimization methods
        
        TarIdxOptUnion <- sort(unique(unlist(TarIdxOpt))) # union target index vector in TargetIdxList over all optimization methods
        TarIdxOptIdx <- lapply(TarIdxOpt, FUN = function(x){match(x, TarIdxOptUnion)}) # list of index of optimal target vector in TarIdxOptUnion, each cell of which corresponding to a task    
        TarOptUnion <- array(TargetAll[TarIdxOptUnion,,], dim = c(length(TarIdxOptUnion), 2, 2)) # union targets selected by last section for the purpose of smoothing, indexed by TarIdxOptUnion
        KOptMax <- max(KOpt) # max K among all tasks
        NeighborPairsUnion <- NeighborGen(Coordinate = Coordinate, SpaSepType = TargetType, SpaSep = TarOptUnion) # generate set of curve IDs for each neighborhood target
        Curves <- 1 : NCurve
        SmoothList <- SmoothCore(DataMat = DataMat, Curves = Curves, NeighborPairs = NeighborPairsUnion, EvalTime.Mu = EvalTime.Mu, K = KOptMax, 
                                 EvalTime.Diag = EvalTime.Diag, Rotate = Rotate, LLQParams = LLQParams, SigmaMethod = SigmaMethod, BWMethod = BWMethod, 
                                 DF = DF, Verbose = Verbose) 
        # the comprehensive smoothing result      
        LambdaEst <- SmoothList$LambdaEst # depending on K
        XLambdaEst <- SmoothList$XLambdaEst # depending on K and NTargets
        EigenFunEst <- SmoothList$EigenFunEst # depending on K 
        XEigenFunEst <- SmoothList$XEigenFunEst # depending on K and NTargets
        CovFitMat <- SmoothList$CovFitMat # estiamted covariance surface
        XCovFitMat <- SmoothList$XCovFitMat # estiamted cross-covariance surface
        FitMu.Eval <- SmoothList$FitMu.Eval # estiamted mean
        FitMu <- SmoothList$FitMu # estimated mean stacked across all curves
        StepSize <- SmoothList$StepSize # step size
        SigmaEst.Rotate <- SmoothList$SigmaEst.Rotate # estimated sigma
        SigmaEst.UnRotate <- SmoothList$SigmaEst.UnRotate # estimated sigma
        
      # estiamte correlations from individual eigenvalues
        
        NTarOptUnion <- length(TarIdxOptUnion)
        RhoEst.Ind <- XLambdaEst / RepMat(t(LambdaEst), NTarOptUnion, 1) # ratio of individual eigenvalues from target union. A matrix of size NTarOptUnion by KOptMax
        
      # fit matern model parameters
        
        if(ModelSelection) # If ModelSelection, then each optimization method would have its own specific optimal Target and K. Then, optimization has to be done one task by another
        {
          RhoEst.Proj <- list()        
          NTask <- length(Optim.Method) 
          ParFitted.Ind <- list()
          Converge.Ind <- list()
          ParFittedPull.Ind <- matrix(NA, NTask, 4)
          ConvergePull.Ind <- matrix(NA, NTask, 1)          
          ParFitted.Proj <- list()
          Converge.Proj <- list()
          ParFittedPull.Proj <- matrix(NA, NTask, 4)
          ConvergePull.Proj <- matrix(NA, NTask, 1)        
        
          for(task in 1 : NTask)
          {
            ThisTarIdxOpt <- TarIdxOptUnion[TarIdxOptIdx[[task]]] # current optimal target index vector
            ThisTarOpt <- array(TargetAll[ThisTarIdxOpt,,], dim = c(length(ThisTarIdxOpt), dim(TargetAll)[2], dim(TargetAll)[3])) # current optimal target 
            ThisTarRepOpt <- TargetRep[ThisTarIdxOpt,] # current optimal target representative
            if(is.vector(ThisTarRepOpt)){ThisTarRepOpt <- rbind(ThisTarRepOpt)}
            ThisKOpt <- KOpt[task] # current optimal K
            ThisProjList <- RhoEstProj(CovFitMat, XCovFitMat[ThisTarIdxOpt,,], K = ThisKOpt) # current list of projected estimation of eigenstructure
            RhoEst.Proj[[task]] <- ThisProjList$RhoEst.Proj # ratio of projected eigenvalues, a matrix of size NTarget by K for each task where NTarget maybe different
            ThisDefaultPars <- rbind(DefaultPars[task,])
            ThisInitialPars <- rbind(InitialPars[task,])
            ThisParsBound <- array(ParsBound[task,,], dim = c(1, dim(ParsBound)[2:3]))
            ThisOptim.Method <- Optim.Method[task]
            ThisNGridsSearch <- rbind(NGridsSearch[task,])
            ThisLog <- rbind(Log[task,])
            ThisMaxIt <- MaxIt[task]
            ParFitted.Ind[[task]] <- matrix(NA, ThisKOpt, 4)
            Converge.Ind[[task]] <- rep(NA, ThisKOpt)
            ParFitted.Proj[[task]] <- matrix(NA, ThisKOpt, 4)
            Converge.Proj[[task]] <- rep(NA, ThisKOpt)          
            
            for(k in 1 : ThisKOpt)
            {
              thisRho <- RhoEst.Ind[TarIdxOptIdx[[task]],k] # fit correlation model based on rho estimated from Kth individual eigenvalues
              ResultList <- NULL
              try(ResultList <- AnisoMaternFit(Estimates = thisRho, DefaultPars = ThisDefaultPars, InitialPars = ThisInitialPars, 
                                               ParsBound = ThisParsBound, TargetRep = ThisTarRepOpt, TargetType = TargetType, 
                                               Optim.Method = ThisOptim.Method, Log = ThisLog, MaxIt = ThisMaxIt, 
                                               NGridsSearch = ThisNGridsSearch), silent = TRUE)
              if(!is.null(ResultList))
              {
                ParFitted.Ind[[task]][k,] <- ResultList$ParFitted
                Converge.Ind[[task]][k] <- ResultList$Converge
              }
              
              thisRho <- RhoEst.Proj[[task]][,k] # fit correlation model based on rho estimated from Kth projected eigenvalues
              ResultList <- NULL
              try(ResultList <- AnisoMaternFit(Estimates = thisRho, DefaultPars = ThisDefaultPars, InitialPars = ThisInitialPars, 
                                               ParsBound = ThisParsBound, TargetRep = ThisTarRepOpt, TargetType = TargetType, 
                                               Optim.Method = ThisOptim.Method, Log = ThisLog, MaxIt = ThisMaxIt, 
                                               NGridsSearch = ThisNGridsSearch), silent = TRUE)
              if(!is.null(ResultList))
              {
                ParFitted.Proj[[task]][k,] <- ResultList$ParFitted
                Converge.Proj[[task]][k] <- ResultList$Converge
              }
            }     
            
            ResModelK <- tolower(ResModelK)
            if(ResModelK != "no")
            {
              if(ResModelK == "opt")
              {
                thisPullK <- ThisKOpt
              }else
              {
                thisPullK <- as.numeric(ResModelK)
              }
              thisPullRho <- as.vector(RhoEst.Ind[TarIdxOptIdx[[task]], 1:thisPullK]) # fit correlation model based on pulled rho estimated from individual eigenvalues
              ThisPullTarRepOpt <- RepMat(rbind(ThisTarRepOpt), thisPullK, 1)
              ResultList <- NULL
              try(ResultList <- AnisoMaternFit(Estimates = thisPullRho, DefaultPars = ThisDefaultPars, InitialPars = ThisInitialPars, 
                                               ParsBound = ThisParsBound, TargetRep = ThisPullTarRepOpt, TargetType = TargetType, 
                                               Optim.Method = ThisOptim.Method, Log = ThisLog, MaxIt = ThisMaxIt, 
                                               NGridsSearch = ThisNGridsSearch), silent = TRUE)
              if(!is.null(ResultList))
              {
                ParFittedPull.Ind[task,] <- ResultList$ParFitted
                ConvergePull.Ind[task] <- ResultList$Converge
              }
              
              thisPullRho <- as.vector(RhoEst.Proj[[task]][, 1:thisPullK]) # fit correlation model based on pulled rho estimated from projected eigenvalues
              ThisPullTarRepOpt <- RepMat(rbind(ThisTarRepOpt), thisPullK, 1)
              ResultList <- NULL
              try(ResultList <- AnisoMaternFit(Estimates = thisPullRho, DefaultPars = ThisDefaultPars, InitialPars = ThisInitialPars, 
                                               ParsBound = ThisParsBound, TargetRep = ThisPullTarRepOpt, TargetType = TargetType, 
                                               Optim.Method = ThisOptim.Method, Log = ThisLog, MaxIt = ThisMaxIt, 
                                               NGridsSearch = ThisNGridsSearch), silent = TRUE)
              if(!is.null(ResultList))
              {
                ParFittedPull.Proj[task,] <- ResultList$ParFitted
                ConvergePull.Proj[task] <- ResultList$Converge
              }         
            }
          }
        }else # if no ModelSelection, then an expanding number of target index are fitted. For each given target index vector, all optimization methods can be fitted in one function
        { 
          NTargetAll <- dim(TargetAll)[1]
          NTask <- length(Optim.Method) 
          RhoEst.Proj <- list()   
          
          ParFitted.Ind <- array(NA, dim = c(NTargetAll, KOptMax, NTask, 4))
          ParFittedPull.Ind <- array(NA, dim = c(NTargetAll, NTask, 4))
          Converge.Ind <- array(NA, dim = c(NTargetAll, KOptMax, NTask))
          ConvergePull.Ind <- array(NA, dim = c(NTargetAll, NTask))
          
          ParFitted.Proj <- array(NA, dim = c(NTargetAll, KOptMax, NTask, 4))
          ParFittedPull.Proj <- array(NA, dim = c(NTargetAll, NTask, 4))                                     
          Converge.Proj <- array(NA, dim = c(NTargetAll, KOptMax, NTask))
          ConvergePull.Proj <- array(NA, dim = c(NTargetAll, NTask))
          
          for(t in 1 : NTargetAll)
          {            
            RhoEst.Proj[[t]] <- RhoEstProj(CovFitMat, XCovFitMat[1:t,,], K = KOptMax)$RhoEst.Proj
            TarRep <- TargetRep[1:t,]
            if(is.vector(TarRep)){TarRep <- rbind(TarRep)}
            for(k in 1 : KOptMax)
            {
              thisRho <- RhoEst.Ind[1:t,k]
              ResultList <- NULL
              try(ResultList <- AnisoMaternFit(Estimates = thisRho, DefaultPars = DefaultPars, InitialPars = InitialPars, 
                                               ParsBound = ParsBound, TargetRep = TarRep, TargetType = TargetType, 
                                               Optim.Method = Optim.Method, Log = Log, MaxIt = MaxIt, 
                                               NGridsSearch = NGridsSearch), silent = TRUE)
              if(!is.null(ResultList))
              {
                ParFitted.Ind[t,k,,] <- ResultList$ParFitted
                Converge.Ind[t,k,] <- ResultList$Converge
              }
              
              thisRho <- RhoEst.Proj[[t]][,k] # fit correlation model with 
              ResultList <- NULL
              try(ResultList <- AnisoMaternFit(Estimates = thisRho, DefaultPars = DefaultPars, InitialPars = InitialPars, 
                                               ParsBound = ParsBound, TargetRep = TarRep, TargetType = TargetType, 
                                               Optim.Method = Optim.Method, Log = Log, MaxIt = MaxIt, 
                                               NGridsSearch = NGridsSearch), silent = TRUE)
              if(!is.null(ResultList))
              {
                ParFitted.Proj[t,k,,] <- ResultList$ParFitted
                Converge.Proj[t,k,] <- ResultList$Converge
              }
            }
            
            ResModelK <- tolower(ResModelK)
            if(ResModelK != "no") #
            {
              if(ResModelK == "opt")
              {
                thisPullK <- ThisKOpt
              }else
              {
                thisPullK <- as.numeric(ResModelK)
              }
              thisPullRho <- as.vector(RhoEst.Ind[1:t, 1:thisPullK]) # fit correlation model based on pulled rho estimated from individual eigenvalues
              ThisPullTarRepOpt <- RepMat(rbind(TarRep), thisPullK, 1)
              ResultList <- NULL
              try(ResultList <- AnisoMaternFit(Estimates = thisPullRho, DefaultPars = DefaultPars, InitialPars = InitialPars, 
                                               ParsBound = ParsBound, TargetRep = ThisPullTarRepOpt, TargetType = TargetType, 
                                               Optim.Method = Optim.Method, Log = Log, MaxIt = MaxIt, 
                                               NGridsSearch = NGridsSearch), silent = TRUE)
              if(!is.null(ResultList))
              {
                ParFittedPull.Ind[t,,] <- ResultList$ParFitted
                ConvergePull.Ind[t,] <- ResultList$Converge
              }
              
              thisPullRho <- as.vector(RhoEst.Proj[[t]][, 1:thisPullK]) # fit correlation model based on pulled rho estimated from projected eigenvalues
              ThisPullTarRepOpt <- RepMat(rbind(TarRep), thisPullK, 1)
              ResultList <- NULL
              try(ResultList <- AnisoMaternFit(Estimates = thisPullRho, DefaultPars = DefaultPars, InitialPars = InitialPars, 
                                               ParsBound = ParsBound, TargetRep = ThisPullTarRepOpt, TargetType = TargetType, 
                                               Optim.Method = Optim.Method, Log = Log, MaxIt = MaxIt, 
                                               NGridsSearch = NGridsSearch), silent = TRUE)
              if(!is.null(ResultList))
              {
                ParFittedPull.Proj[t,,] <- ResultList$ParFitted
                ConvergePull.Proj[t,] <- ResultList$Converge
              }         
            }
          }         
        }
        
      # Curve Reconstruction
        
        if(Reconstruction)
        {
          if(ModelSelection)
          {
            Reconstruction <- list()
            for(task in 1 : NTask)
            {
              ThisKOpt <- KOpt[task]
              ThisOmega.Ind <- array(NA, dim = c(NCurve, NCurve, ThisKOpt))
              ThisOmega.Proj <- array(NA, dim = c(NCurve, NCurve, ThisKOpt))
              Omega0 <- array(NA, dim = c(NCurve, NCurve, ThisKOpt))
              for(k in 1 : ThisKOpt)
              {
                ThisPar <- ParFitted.Ind[[task]][k,]
                ThisOmega.Ind[,,k] <- AnisoMatern(Coordinate = Coordinate, Alpha = ThisPar[1], Delta = ThisPar[2], Kappa = ThisPar[3], Phi = ThisPar[4], Log = c(0,0,0,0))
                ThisPar <- ParFitted.Proj[[task]][k,]
                ThisOmega.Proj[,,k] <- AnisoMatern(Coordinate = Coordinate, Alpha = ThisPar[1], Delta = ThisPar[2], Kappa = ThisPar[3], Phi = ThisPar[4], Log = c(0,0,0,0))
                Omega0[,,k] <- diag(1, NCurve, NCurve)
              }
              Reconstruction[[task]]$IndRotate <- ReConstruct(DataMat = DataMat, TimeIndex = TimeIndex, FitMu = FitMu, EigenFun = EigenFunEst, Omega = ThisOmega.Ind, Lambda = LambdaEst, Sigma = SigmaEst.Rotate, FitMu.Regular = FitMu.Eval)
              Reconstruction[[task]]$IndUnRotate <- ReConstruct(DataMat = DataMat, TimeIndex = TimeIndex, FitMu = FitMu, EigenFun = EigenFunEst, Omega = ThisOmega.Ind, Lambda = LambdaEst, Sigma = SigmaEst.UnRotate, FitMu.Regular = FitMu.Eval)
              Reconstruction[[task]]$ProjRotate <- ReConstruct(DataMat = DataMat, TimeIndex = TimeIndex, FitMu = FitMu, EigenFun = EigenFunEst, Omega = ThisOmega.Proj, Lambda = LambdaEst, Sigma = SigmaEst.Rotate, FitMu.Regular = FitMu.Eval)
              Reconstruction[[task]]$ProjUnRotate <- ReConstruct(DataMat = DataMat, TimeIndex = TimeIndex, FitMu = FitMu, EigenFun = EigenFunEst, Omega = ThisOmega.Proj, Lambda = LambdaEst, Sigma = SigmaEst.UnRotate, FitMu.Regular = FitMu.Eval) 
              Reconstruction[[task]]$NoCorRotate <- ReConstruct(DataMat = DataMat, TimeIndex = TimeIndex, FitMu = FitMu, EigenFun = EigenFunEst, Omega = Omega0, Lambda = LambdaEst, Sigma = SigmaEst.Rotate, FitMu.Regular = FitMu.Eval) 
              Reconstruction[[task]]$NoCorUnRotate <- ReConstruct(DataMat = DataMat, TimeIndex = TimeIndex, FitMu = FitMu, EigenFun = EigenFunEst, Omega = Omega0, Lambda = LambdaEst, Sigma = SigmaEst.UnRotate, FitMu.Regular = FitMu.Eval) 
            }
          }else
          {
            Reconstruction <- array(list(), dim = c(NTask, NTargetAll))
            Omega0 <- array(NA, dim = c(NCurve, NCurve, KOptMax))
            for(task in 1 : NTask)
            {
              for(t in 1 : NTargetAll)
              {
                ThisOmega.Ind <- array(NA, dim = c(NCurve, NCurve, KOptMax))
                ThisOmega.Proj <- array(NA, dim = c(NCurve, NCurve, KOptMax))
                for(k in 1 : KOptMax)
                {
                  ThisPar <- ParFitted.Ind[t,k,task,]
                  ThisOmega.Ind[,,k] <- AnisoMatern(Coordinate = Coordinate, Alpha = ThisPar[1], Delta = ThisPar[2], Kappa = ThisPar[3], Phi = ThisPar[4], Log = c(0,0,0,0))
                  ThisPar <- ParFitted.Proj[t,k,task,]
                  ThisOmega.Proj[,,k] <- AnisoMatern(Coordinate = Coordinate, Alpha = ThisPar[1], Delta = ThisPar[2], Kappa = ThisPar[3], Phi = ThisPar[4], Log = c(0,0,0,0))
                  Omega0[,,k] <- diag(1, NCurve, NCurve)
                }
                Reconstruction[[task, t]]$IndRotate <- ReConstruct(DataMat = DataMat, TimeIndex = TimeIndex, FitMu = FitMu, EigenFun = EigenFunEst, Omega = ThisOmega.Ind, Lambda = LambdaEst, Sigma = SigmaEst.Rotate, FitMu.Regular = FitMu.Eval)
                Reconstruction[[task, t]]$IndUnRotate <- ReConstruct(DataMat = DataMat, TimeIndex = TimeIndex, FitMu = FitMu, EigenFun = EigenFunEst, Omega = ThisOmega.Ind, Lambda = LambdaEst, Sigma = SigmaEst.UnRotate, FitMu.Regular = FitMu.Eval)
                Reconstruction[[task, t]]$ProjRotate <- ReConstruct(DataMat = DataMat, TimeIndex = TimeIndex, FitMu = FitMu, EigenFun = EigenFunEst, Omega = ThisOmega.Proj, Lambda = LambdaEst, Sigma = SigmaEst.Rotate, FitMu.Regular = FitMu.Eval)
                Reconstruction[[task, t]]$ProjUnRotate <- ReConstruct(DataMat = DataMat, TimeIndex = TimeIndex, FitMu = FitMu, EigenFun = EigenFunEst, Omega = ThisOmega.Proj, Lambda = LambdaEst, Sigma = SigmaEst.UnRotate, FitMu.Regular = FitMu.Eval) 
                Reconstruction[[task, t]]$NoCorRotate <- ReConstruct(DataMat = DataMat, TimeIndex = TimeIndex, FitMu = FitMu, EigenFun = EigenFunEst, Omega = Omega0, Lambda = LambdaEst, Sigma = SigmaEst.Rotate, FitMu.Regular = FitMu.Eval) 
                Reconstruction[[task, t]]$NoCorUnRotate <- ReConstruct(DataMat = DataMat, TimeIndex = TimeIndex, FitMu = FitMu, EigenFun = EigenFunEst, Omega = Omega0, Lambda = LambdaEst, Sigma = SigmaEst.UnRotate, FitMu.Regular = FitMu.Eval) 
              }
            }  
          }
        }else
        {
          Reconstruction <- NULL
        }
        
      # return results
        
        results <- list(ParFitted.Ind = ParFitted.Ind, ParFitted.Proj = ParFitted.Proj, FitMu.Eval = FitMu.Eval, RhoEst.Ind = RhoEst.Ind, 
                        Converge.Ind = Converge.Ind, RhoEst.Proj = RhoEst.Proj, Converge.Proj = Converge.Proj, SigmaEst.Rotate = SigmaEst.Rotate, 
                        SigmaEst.UnRotate = SigmaEst.UnRotate, SmoothList = SmoothList, ModelSelList = ModelSelList, ParFittedPull.Ind = ParFittedPull.Ind, 
                        ParFittedPull.Proj = ParFittedPull.Proj, ConvergePull.Ind = ConvergePull.Ind, ConvergePull.Proj = ConvergePull.Proj, DataMat = DataMat, 
                        TimeIndex = TimeIndex, Reconstruction = Reconstruction)
        return(results)
        
    }