  
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  ########################################################################
  ####    Function to Smooth Mean, Covariance and XCovariance with    ####
  ########################################################################
  ####        version 1.0 Created by Chong Liu on 2012/11/14          ####
  ########################################################################
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#                                              
  # Input:
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.DataMat:          A matrix of size NAllObs by 3, with the first column being curve IDs, the second column being the observations Y and the third colum being corresponding time
  #                     points.
  # 2.Curves:           A vector of curve IDs indicate which curves are used in this smoothing
  # 3.NeighborPairs:    A list of length NTarget in which each cell is a matrix of size Npairs by 2, of all pair of neighboring locations under the current target
  # 4.EvalTime.Mu:      A vector of length NGrid which is the time points used to evaluate fitted mean
  # 5.EvalTime2.Mu:     An optional vector of time points used to evaluate fitted mean, in case of other purposes
  # 6.K:                Number of eigenfunctions retained in results.
  # 7.EvalTime.Diag:    A vector of time points used to evaluate diagonal component of the covariance surface
  # 8.Rotate:           A logical value indicates whether to rotate axes and use local linear-quadratic smoothing for the off-diagonal component of the covariance surface 
  #                     If Rotate = 2, then it means both rotated and unrotated results are computed
  # 9.LLQParams:        A list of parameters for local linear quadratic smoothing, including
  #                     Method:    Methods used for local linear-quadratic smoothing
  #                                "poly" uses (1,X1-X0,X2-X0,(X2-X0)^2 and MA750 type of polynomial expressions to solve for the local fitting
  #                                "matrix" uses (1,X1,X2,X2^2) and compact matrix expressions to solve for the local fitting
  #                     InvSym:    A facas object in Ryacas package of matrix inverse used for corresponding method indicated by Method. Reason we need this argument is symbolic 
  #                                inverse of 4 by 4 matrix takes a while. In simulations, this symbolic inverse is the same for each simulated data but evaluated at different points. 
  #                                It is time consuming to do this inverse each time. Instead, it can be provided from outside and we only evaluate the inverse object at different points
  #                                using this function in simulations.   
  # 10.SigmaMethod:     A vector of methods used for estimating Sigma
  #                     "PACE", the original PACE method 
  #                     "Modified PACE", the modified PACE method
  #                     "Eigenvalue", method where taking the eigenvalue difference between using and not using diagonal variance component.
  # 11.BWMethod:        Bandwidth method
  # 12.DF:              Degree of freedom used in DF method in choosing bandwidth
  # 13.Verbose:         Logical value indicating whether each neighbor pair is finished
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Output
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.LambdaEst:        
  # 2.EigenFunEst:
  # 3.XLambdaEst:
  # 4.XEigenFunEst:
  # 5.FitMu:
  # 6.FitMu.Eval:
  # 7.FitMu.Eval2:
  # 8.CovFitMat:
  # 9.StepSize:
  # 10.SigmaEst:
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  
  SmoothCore <- function(DataMat, Curves, NeighborPairs, EvalTime.Mu, EvalTime2.Mu = NULL, K, EvalTime.Diag, Rotate, LLQParams = NULL, SigmaMethod = "PACE", BWMethod, DF = NA, Verbose = TRUE)
  {
    
    # load packages and functions
    
    # library(fda)
    # library(sm)
    # source("LocLinReg.r")
    # source("SmoothCov.r")
    # source("SmoothXCov.r")
    
    # prepare variables
    
      ID <- as.matrix(DataMat[,1])
      Idx <- is.element(ID, Curves)
      Y <- as.matrix(DataMat[,2])
      Time <- as.matrix(DataMat[,3])
    
    # fit mean
    
      HMu <- HSelect(X = as.matrix(Time[Idx]), Y = as.matrix(Y[Idx]), BWMethod = BWMethod, LOPO = 2, LocLinQua = FALSE, NbinsX = -1, DF = DF)    		
      FitMu.Full <- LocLinReg(X = as.matrix(Time[Idx]), Y = as.matrix(Y[Idx]), X0 = as.matrix(Time), H = HMu, NbinsX = -1, NbinsX0 = 0)$M # evaluation at all time points in DataMat
      FitMu <- FitMu.Full[Idx] # evaluation at training time points
      FitMu.Eval <- LocLinReg(X = as.matrix(Time[Idx]), Y = as.matrix(Y[Idx]), X0 = as.matrix(EvalTime.Mu), H = HMu, NbinsX = -1, NbinsX0 = 0)$M # evaluation at separate testing time points (regular grids)
      if(!is.null(EvalTime2.Mu))
      {
        FitMu.Eval2 <- LocLinReg(X = as.matrix(Time[Idx]), Y = as.matrix(Y[Idx]), X0 = as.matrix(EvalTime2.Mu), H = HMu, NbinsX = -1, NbinsX0 = 0)$M # evaluation at second separate testing time points (arbitrary testing points)
      }else
      {
        FitMu.Eval2 <- NULL 
      }
      
    # smooth the covariance surface
    
      NGrid <- length(EvalTime.Mu)
      EvalGrid <- cbind(rep(EvalTime.Mu, time = NGrid), rep(EvalTime.Mu, each = NGrid))
      CovFitList <- SmoothCov(Curves = Curves, DataMat = DataMat, K = K, FitMu = FitMu.Full, LLQParams = LLQParams, EvalGrid = EvalGrid, VarEvalTime = EvalTime.Diag, Structure.2D = "common", Rotate = Rotate, 
                              SigmaMethod = SigmaMethod, BWMethod = BWMethod, DF = DF)      
      StepSize <- EvalTime.Mu[2] - EvalTime.Mu[1] 
      SigmaEst.Rotate <- sqrt(CovFitList$SigmaSq$PACE.Rotate)
      SigmaEst.UnRotate <- sqrt(CovFitList$SigmaSq$PACE.UnRotate)
      LambdaEst <- CovFitList$EigenVal * StepSize # normalize for discrete evaluation, a vector of length KMax
      EigenFunEst <- CovFitList$EigenFun / sqrt(StepSize) # normalize the continuous function to have norm 1, a matrix of size NGrid by KMax
      CovFitMat <- CovFitList$CovFitMat

    # smooth the Xcovariance surface for optimal number of Targets selected
    
      XCovFitList <- SmoothXCov(NeighborPairs = NeighborPairs, DataMat = DataMat, K = K, FitMu = FitMu.Full, EvalGrid = EvalGrid, Structure.2D = "common", BWMethod = BWMethod, DF = DF, Verbose = Verbose)      
      XLambdaEst <- XCovFitList$EigenVal * StepSize
      XEigenFunEst <- XCovFitList$EigenFun / sqrt(StepSize)
      XCovFitMat <- XCovFitList$XCovFitMat
        
    # return results
      
      return(list(LambdaEst = LambdaEst, EigenFunEst = EigenFunEst, XLambdaEst = XLambdaEst, XEigenFunEst = XEigenFunEst, FitMu = FitMu, FitMu.Eval = FitMu.Eval, 
                  FitMu.Eval2 = FitMu.Eval2, CovFitMat = CovFitMat, XCovFitMat = XCovFitMat, StepSize = StepSize, SigmaEst.Rotate = SigmaEst.Rotate, 
                  SigmaEst.UnRotate = SigmaEst.UnRotate))
  }
          
  
  
  
  
  