  
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  ################################################################
  ####    Function of Local Linear Smoothing of COvariance    ####
  ################################################################
  ####     version 2.0 Created by Chong Liu on 2013/01/04     ####
  ################################################################
  #### change from version 1.0                                ####
  #### 1.add eigenvalue method of estimating sigma            ####
  #### 2.remove RmDiag                                        ####
  ################################################################
  #### change of version 3.0 from version 2.0                 ####
  #### 1.change h.select to HSelect                           ####
  ################################################################
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#                                              
  # Input:
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.Curves:           A vector of curve IDs indicate which curves are used in this smoothing
  # 2.DataMat:          A matrix of size NAllObs by 3, with the first column being curve IDs, the second column being the observations Y and the third colum being corresponding time 
  #                     points. 
  # 3.K:                Number of eigenfunctions retained in results.
  # 4.FitMu:            Fitted mean curve evaluated at all observed time points stacked together in column.
  # 5.EvalGrid:         A matrix of size NGrid^2 by 2 which are grid points used to evaluate the covariance surface, required to have same time points on each direction.
  # 6.VarEvalTime:      A vector of time points used to evaluate diagonal component of the covariance surface
  # 7.Structure.2D:     Type of bandwidth in two-dimension surface smoothing
  #                     "common", equal bandwidth for two dimensions;
  #                     "scale", bandwidths proportional to dimension standard deviation;
  #                     "separate", arbitrary bandwidths for the two dimensions.
  # 8.Rotate:           A logical value indicates whether to rotate axes and use local linear-quadratic smoothing for the off-diagonal component of the covariance surface 
  #                     If Rotate = 2, then it means both rotated and unrotated results are computed
  # 9.LLQParams:        A list of parameters for local linear quadratic smoothing, including
  #                     Method:    Methods used for local linear-quadratic smoothing
  #                                "poly" uses (1,X1-X0,X2-X0,(X2-X0)^2 and MA750 type of polynomial expressions to solve for the local fitting
  #                                "matrix" uses (1,X1,X2,X2^2) and compact matrix expressions to solve for the local fitting
  #                     InvSym:    A facas object in Ryacas package of matrix inverse used for corresponding method indicated by Method. Reason we need this argument is symbolic 
  #                                inverse of 4 by 4 matrix takes a while. In simulations, this symbolic inverse is the same for each simulated data but evaluated at different points. 
  #                                It is time consuming to do this inverse each time. Instead, it can be provided from outside and we only evaluate the inverse object at different points
  #                                using this function in simulations.  
  # 10.SigmaMethod:     A vector of methods used for estimating Sigma
  #                     "PACE", the original PACE method 
  #                     "Modified PACE", the modified PACE method
  #                     "Eigenvalue", method where taking the eigenvalue difference between using and not using diagonal variance component.
  # 11.BWMethod:        Bandwidth method
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Output
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.BW:               A list which contains selected bandwidths for different smoothings
  #                     BW$Cov.Off, the 2-dim bandwidths selected for smoothing on off diagonal component of the empirical covariance 
  #                     BW$Cov.All, the 2-dim bandwidths selected for smoothing on the entire empirical covariance
  #                     BW$Diag, the 1-dim bandwidth selected for the diagonal component of the empirical covariance
  # 2.CovFitMat:        A matrix of size of NGrid by NGrid which stores fitted Covs evaluated at EvalGrid
  # 3.EigenVal:         A vector of length K which stores the first K eigenvalues of CovFitMat
  # 4.EigenFun:         A matrix of size NGrid by K which stores the first K eigenfunctions of CovFitMat, evaluated at time points used to construct the EvalGrid 
  # 5.VarFit1D:         A vector of length VarEvalTime which is the 1-dim smoothed diagonal of the empirical covariance,(supposed to have the noise variance component)
  # 6.VarFit2D:         A vector of length VarEvalTime which is the evaluation at VarEvalTime of the 2-dim smoothing of the off diagonal component of the empirical covariance
  # 7.SigmaSq:          A list of estimated variance of the noise
  #                     SigmaSq$EigAll, derived from the eigenvalues of the entire empirical surface using BW$Cov.All
  #                     SigmaSq$EigOff, derived from the eigenvalues of the entire empirical surface using BW$Cov.Off
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#

  SmoothCov <- function(Curves, DataMat, K, FitMu, EvalGrid, VarEvalTime, Structure.2D, Rotate, LLQParams = NULL, SigmaMethod = "PACE", BWMethod, DF = NA)
  {
    # load packages and functions
    
      #source("LocLinReg.r")
      #source("LocLinQuaReg2D")
   
    # prepare data for fitting 
      
      NCurve <- length(Curves) # number of curves used for smoothing
      ID <- DataMat[,1] # curve IDs corresponding to each observation
      NObs <- as.vector(table(ID)) # number of observations on each curve
      nnAll  <- sum(NObs[Curves]^2) # total number of empirical covariances
      PreFit <- matrix(NA, nnAll, 3) # matrix which stores all empirical covariances
      DataMatPlus <- cbind(DataMat, FitMu) 
      StepSize <- diff(unique(as.vector(EvalGrid)))[1]
      BW <- list()
      
      count <- 0
      for(i in 1 : NCurve)
      {
        ThisCurve <- Curves[i] # current curve
        chunck <- DataMatPlus[DataMatPlus[,1] == ThisCurve, ] # extract the chunck of data corresponding to the current curve
        if(is.vector(chunck)){chunck <- rbind(chunck)}
        n <- dim(chunck)[1] # should be equal to NObs[ThisCurve] for double check
        tempidx <- (count + 1) : (count + n^2) 
        PreFit[tempidx, 1] <- rep(chunck[,3], times = n) # the first col is the first time point in time pairs
        PreFit[tempidx, 2] <- rep(chunck[,3], each = n) # the second col is the second time point in time pairs
        PreFit[tempidx, 3] <- as.vector((chunck[,2] - chunck[,4]) %o% (chunck[,2] - chunck[,4])) # all empirical covariances from the current curve
        count <- count + n^2
      }   
      
    # construct off-diagonal (Covariance) and diagonal (Variance) components of var-covariance surface   
      
      DiagComp <- PreFit[PreFit[,1] == PreFit[,2],] # extract rows corresponding variances and extract all three columns (The first two are identical time points with the third the empirical variance)
      DiagComp[,2] <- DiagComp[,2] * sqrt(2) # rescale real time (the second column) on the diagonal to get rotated time for the 1-dim smoothing; the diagonal component must be rotated for 1-dim smoothing
      OffDiagComp <- PreFit[PreFit[,1] != PreFit[,2],] # remove rows corresponding to covariances, where the first time equal the second time
      AllComp <- PreFit
      RotateMat <- matrix(c(sqrt(2)/2, sqrt(2)/2, -sqrt(2)/2, sqrt(2)/2), 2, 2, byrow = TRUE) # rotate matrix for coordinate system (s,t)
      
      if(Rotate) # rotate the axes if Rotate is true, applied only to the off-diagonal component
      {
        OffDiagComp[,c(1,2)] <- OffDiagComp[,c(1,2)] %*% t(RotateMat) # rescale real time to get rotated time for off-diagonal 2-dim smoothing
        AllComp[,c(1,2)] <- AllComp[,c(1,2)] %*% t(RotateMat) # rescale real time to get rotated time for entire component 2-dim smoothing
        EvalGrid <- EvalGrid %*% t(RotateMat)
        DesignMat.UnRotate <- OffDiagComp[,c(1,2)] %*% solve(t(RotateMat)) 
        DesignMat.Rotate <- OffDiagComp[,c(1,2)]
      }else
      {
        DesignMat.Rotate <- OffDiagComp[,c(1,2)] %*% t(RotateMat) 
        DesignMat.UnRotate <- OffDiagComp[,c(1,2)]          
      } 
      
    # smooth off-diagonal component of the var-covariance surface      

      HCov.OffDiag.Rotate <- HSelect(X = DesignMat.Rotate, Y = as.matrix(OffDiagComp[,3]), LOPO = 2, Structure.2D = Structure.2D, LocLinQua = TRUE, LLQParams = LLQParams, NbinsX = -1, BWMethod = BWMethod, DF = DF) # select bandwidths as for 2-dim local linear smoothing on X1 and X2^2      
      HCov.OffDiag.UnRotate <- HSelect(X = DesignMat.UnRotate, Y = as.matrix(OffDiagComp[,3]), LOPO = 2, Structure.2D = Structure.2D, LocLinQua = FALSE, NbinsX = -1, BWMethod = BWMethod, DF = DF)# no need to provide LLQParams since LocLinQua = FALSE      
      
      if(Rotate)
      { 
        CovFitVec <- LocLinQuaReg2D(X = OffDiagComp[,c(1,2)], Y = as.matrix(OffDiagComp[,3]), X0 = EvalGrid, H = HCov.OffDiag.Rotate, NbinsX = -1, NbinsX0 = 0, Method = LLQParams$Method, InvSym = LLQParams$InvSym)$M # fit, with rotated axes, the off-diagonal component of the covariance surface in vector format      
        BW$Cov.Off <- HCov.OffDiag.Rotate
      }else
      {
        CovFitVec <- LocLinReg(X = OffDiagComp[,c(1,2)], Y = as.matrix(OffDiagComp[,3]), X0 = EvalGrid, H = HCov.OffDiag.UnRotate, NbinsX = -1, NbinsX0 = 0)$M # fit, without rotating axes, the off-diagonal component of the covariance surface in vector format 
        BW$Cov.Off <- HCov.OffDiag.UnRotate
      }  
      if(any(is.nan(CovFitVec))){warnings("bandwidths are too small")}
      
    # compute eigenstructure based on off-diagonal component smoothing
      
      NGrid <- dim(EvalGrid)[1]^0.5 # number of evaluation time points on each direction
      covfitmat <- matrix(CovFitVec, NGrid, NGrid, byrow = FALSE) # evaluated raw covariance matrix at EvalGrid
      covfitmat <- (covfitmat + t(covfitmat)) / 2 # make it symmetric
      eigs <- eigen(covfitmat)
      eigs$values[eigs$values < 1e-8] <- 0 # make negative and small positive eigenvalues to be zeros
      CovFitMat <- eigs$vectors %*% diag(eigs$values) %*% t(eigs$vectors) # reconstruct XCov matrix based on modified eigenvalues      
      EigenVal <- eigs$values[1:K] # store the first K eigenvalues
      EigenFun <- eigs$vectors[,1:K] # store the first K eigenfunctions
     
    # estimate sigma^2
      
      SigmaSq <- list()
      
      if(is.element("Eigenvalue", SigmaMethod))
      {
        if(Rotate)
        {
          HCov.All <- HSelect(X = AllComp[,1:2], y = as.matrix(AllComp[,3]), LOPO = 2, Structure.2D = Structure.2D, LocLinQua = TRUE, LLQParams = LLQParams, NbinsX = -1, BWMethod = BWMethod, DF = DF) # select bandwidths using HSelect with cv method     
          BW$Cov.All <- HCov.All
          AllFitVec.All <- LocLinQuaReg2D(X = AllComp[,c(1,2)], Y = as.matrix(AllComp[,3]), X0 = EvalGrid, H = HCov.All, NbinsX = -1, NbinsX0 = 0, Method = LLQParams$Method, InvSym = LLQParams$InvSym)$M # fit, with rotated axes, the off-diagonal component of the covariance surface in vector format      
          AllFitVec.Off <- LocLinQuaReg2D(X = AllComp[,c(1,2)], Y = as.matrix(AllComp[,3]), X0 = EvalGrid, H = HCov.OffDiag.Rotate, NbinsX = -1, NbinsX0 = 0, Method = LLQParams$Method, InvSym = LLQParams$InvSym)$M # fit, with rotated axes, the off-diagonal component of the covariance surface in vector format      
        }else
        {
          HCov.All <- HSelect(X = AllComp[,1:2], y = as.matrix(AllComp[,3]), LOPO = 2, Structure.2D = Structure.2D, LocLinQua = FALSE, NbinsX = -1, BWMethod = BWMethod, DF = DF) # select bandwidths using HSelect with cv method     
          BW$Cov.All <- HCov.All
          AllFitVec.All <- LocLinReg(X = OffDiagComp[,c(1,2)], Y = as.matrix(OffDiagComp[,3]), X0 = EvalGrid, H = HCov.All, NbinsX = -1, NbinsX0 = 0)$M # fit, without rotating axes, the off-diagonal component of the covariance surface in vector format 
          AllFitVec.Off <- LocLinReg(X = OffDiagComp[,c(1,2)], Y = as.matrix(OffDiagComp[,3]), X0 = EvalGrid, H = HCov.OffDiag.UnRotate, NbinsX = -1, NbinsX0 = 0)$M # fit, without rotating axes, the off-diagonal component of the covariance surface in vector format 
        }
        allfitmat.All <- matrix(AllFitVec.All, NGrid, NGrid, byrow = FALSE) # evaluated raw covariance matrix at EvalGrid
        allfitmat.Off <- matrix(AllFitVec.Off, NGrid, NGrid, byrow = FALSE) # evaluated raw covariance matrix at EvalGrid
        allfitmat.All <- (allfitmat.All + t(allfitmat.All)) / 2 # make it symmetric
        allfitmat.Off <- (allfitmat.Off + t(allfitmat.Off)) / 2 # make it symmetric
        AllEigenVal.All <- eigen(allfitmat.All)$values[1:K]
        AllEigenVal.Off <- eigen(allfitmat.Off)$values[1:K]
        SigmaSq$EigOff <- (AllEigenVal.Off - EigenVal) * StepSize
        SigmaSq$EigAll <- (AllEigenVal.All - EigenVal) * StepSize
      }
      
      if(is.element("PACE", SigmaMethod))
      {
        HVar <- HSelect(X = as.matrix(DiagComp[,2]), Y = as.matrix(DiagComp[,3]), LOPO = 2, LocLinQua = FALSE, NbinsX = -1, BWMethod = BWMethod, DF = DF) # selecting bandwidth for digonal component of the covariance surface
        BW$Diag <- HVar
        VarEvalTime1D <- VarEvalTime * sqrt(2)
        VarFit1D <- LocLinReg(X = as.matrix(DiagComp[,2]), Y = as.matrix(DiagComp[,3]), X0 = as.matrix(VarEvalTime1D), H = HVar, NbinsX = -1, NbinsX0 = 0)$M # fit the diagonal component using local linear smoothing 
        VarEvalTime2D.Rotate <- cbind(VarEvalTime1D, 0) # a matrix of size NVarEvalTime by 2 within rotated axes
        VarFit2D.Rotate <- LocLinQuaReg2D(X = DesignMat.Rotate, Y = as.matrix(OffDiagComp[,3]), X0 = VarEvalTime2D.Rotate, H = HCov.OffDiag.Rotate, NbinsX = -1, NbinsX0 = 0, Method = LLQParams$Method, InvSym = LLQParams$InvSym)$M # evaluate the diagonal from 2D smoothing of the off-diagonal
        VarEvalTime2D.UnRotate <- cbind(VarEvalTime, VarEvalTime) # a matrix of size NVarEvalTime by 2 within unrotated axes
        VarFit2D.UnRotate <- LocLinReg(X = DesignMat.UnRotate, Y = as.matrix(OffDiagComp[,3]), X0 = VarEvalTime2D.UnRotate, H = HCov.OffDiag.UnRotate, NbinsX = -1, NbinsX0 = 0)$M # evaluate the diagonal from 2D smoothing of the off-diagonal         
        SigmaSq$PACE.Rotate <- mean(VarFit1D - VarFit2D.Rotate)      
        SigmaSq$PACE.UnRotate <- mean(VarFit1D - VarFit2D.UnRotate)      
      }

      if(is.element("PACEModified", SigmaMethod))
      {
       print("under construction")
      }      
      
    # return results
    
      results <- list(BW = BW, CovFitMat = CovFitMat, EigenVal = EigenVal, EigenFun = EigenFun, VarFit1D = VarFit1D, VarFit2D.Rotate = VarFit2D.Rotate, 
                      VarFit2D.UnRotate = VarFit2D.UnRotate, SigmaSq = SigmaSq)
      return(results)
  }