  
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  ######################################################################
  ####    Function of Local Linear Smoothing of Cross-COvariance    ####
  ######################################################################
  ####        version 1.0 Created by Chong Liu on 2012/10/31        ####
  ######################################################################
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#                                              
  # Input:
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.NeighborPairs:    A list of length NTarget in which each cell is a matrix of size Npairs by 2, of all pair of neighboring locations under the current target
  # 2.DataMat:          A matrix of size NAllObs by 3, with the first column being curve IDs, the second column being the observations Y and the third colum being corresponding time 
  #                     points. 
  # 3.K:                Number of eigenfunctions retained in results.
  # 4.FitMu:            Fitted mean curve evaluated at all observed time points stacked together in column.
  # 5.EvalGrid:         A matrix of size NGrid^2 by 2 which are grid points used to evaluate the Xcovariance surface, required to have same time points on each direction.
  # 6.Structure.2D:     Type of bandwidth in two-dimension surface smoothing
  #                     "common", equal bandwidth for two dimensions;
  #                     "scale", bandwidths proportional to dimension standard deviation;
  #                     "separate", arbitrary bandwidths for the two dimensions.
  # 7.BWMethod:         "CV", leave-one-point-out is leaving the whole bin out
  #                     "DF", degree of freedom method (sum of diagonal of the Hat matrix), set by DF input, default is 6 for 1-dim and 12 for 2-dim
  #                     "AICC", modified aic method for time series data
  #                     Details are described in SM package, HSelect is created to replicate SM package and extend it to local quadratic smoothing
  # 8.DF:               Degree of freedom number as a user input
  # 9.Verbose:          Logical value indicating whether each neighbor pair is finished
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Output
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.HCov:             A matrix of size NTarget by 2 which stores bandwidths selected by BWMethod in smoothing XCov surface
  # 2.XCovFitMat:       A matrix of size NTarget by NGrid by NGrid which stores fitted XCovs evaluated at EvalGrid
  # 3.EigenVal:         A matrix of size NTarget by K which stores the first K eigenvalues of XCovFitMat
  # 4.EigenFun:         A matrix of size NTarget by NGrid by K which stores the first K eigenfunctions of XCovFitMat, evaluated at time points used to construct the EvalGrid 
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#

  SmoothXCov <- function(NeighborPairs, DataMat, K, FitMu, EvalGrid, Structure.2D, Verbose = FALSE, BWMethod, DF = NA)
  {
    # load packages and functions
    
      #source("LocLinReg.r")
    
    # initiate containers
    
      NTarget <- length(NeighborPairs) # length of the list, the number of XCovs needed to be smoothed
      NGrid <- dim(EvalGrid)[1]^0.5 # number of evaluation time points on each direction
      HCov <- array(NA, dim = c(NTarget, 2)) # bandwidths used to smooth XCov surfaces
      XCovFitMat <- array(NA, dim = c(NTarget, NGrid, NGrid)) # fitted XCov surface evaluated at EvalGrid
      EigenVal <- array(NA, dim = c(NTarget, K)) # the first K eigenvalues of fitted XCov surface
      EigenFun <- array(NA, dim = c(NTarget, NGrid, K)) # the first K eigenfunctions of fitted XCov surface
      
    # smooth all XCov surfaces
      
      for(target in 1 : NTarget) # loop through all target neighborhoods
      {           
        ## extract the current neigborhood
        
          Pairs <- NeighborPairs[[target]] # Pairs is a matrix of size NPairs by 2 which stores all location pairs satisfying the current target
        
        ## construct design matrix for XCov surface smoothing
        
          NPairs <- dim(Pairs)[1] # number of neighboring pairs under the current target
          ID <- DataMat[,1] # curve IDs corresponding to each observation
          NObs <- as.vector(table(ID)) # number of observations on each curve
          nn <- sum(NObs[Pairs[,1]] * NObs[Pairs[,2]]) # total number of empirical Xcovariances      
          XCovPreFit <- matrix(NA, nn, 3)
          DataMatPlus <- cbind(DataMat, FitMu) 
          
          count <- 0
          for(i in 1 : NPairs)
          {
            p1 <- Pairs[i, 1]
            p2 <- Pairs[i, 2]
            chunck1 <- DataMatPlus[DataMatPlus[,1] == p1, ]
            chunck2 <- DataMatPlus[DataMatPlus[,1] == p2, ]
            if(is.vector(chunck1)){chunck1 <- rbind(chunck1)}
            if(is.vector(chunck2)){chunck2 <- rbind(chunck2)}
            n1 <- dim(chunck1)[1]
            n2 <- dim(chunck2)[1] 
            tempidx <- (count + 1) : (count + n1 * n2) 
            XCovPreFit[tempidx, 1] <- rep(chunck1[,3], times = n2)
            XCovPreFit[tempidx, 2] <- rep(chunck2[,3], each = n1)
            XCovPreFit[tempidx, 3] <- as.vector((chunck1[,2] - chunck1[,4]) %o% (chunck2[,2] - chunck2[,4]))
            count <- count + n1 * n2
          }
        
        ## select bandwidth for Xcovariance surface
        
          HCov[target,] <- HSelect(X = XCovPreFit[,c(1,2)], Y = as.matrix(XCovPreFit[,3]), LOPO = 2, Structure.2D = Structure.2D, LocLinQua = FALSE, NbinsX = -1, BWMethod = BWMethod, DF = DF)
        
        ## fit Xcovariance
        
          XCovFitVec <- LocLinReg(X = XCovPreFit[,c(1,2)], Y = as.matrix(XCovPreFit[,3]), X0 = EvalGrid, H = HCov[target,], NbinsX = -1, NbinsX0 = 0)$M # fitted XCovariance surface in vector format
          xcovfitmat <- matrix(XCovFitVec, NGrid, NGrid, byrow = FALSE) # evaluated raw covariance matrix at EvalGrid
          xcovfitmat <- (xcovfitmat + t(xcovfitmat)) / 2 # make it symmetric
          
        ## compute eigenstructure   
        
          eigs <- eigen(xcovfitmat)
          eigs$values[eigs$values < 1e-8] <- 0 # make negative and small positive eigenvalues to be zeros
          XCovFitMat[target,,] <- eigs$vectors %*% diag(eigs$values) %*% t(eigs$vectors) # reconstruct XCov matrix based on modified eigenvalues      
          EigenVal[target,] <- eigs$values[1:K] # store the first K eigenvalues
          EigenFun[target,,] <- eigs$vectors[,1:K] # store the first K eigenfunctions
          
        ## keep track
          
          if(Verbose){print(paste("Target", target, "is done"))}
      }
      
    # return results
    
      return(list(HCov = HCov, XCovFitMat = XCovFitMat, EigenVal = EigenVal, EigenFun = EigenFun))		
    
  }