  
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  #
  ####################################################################
  ####    Function of Computing Matrix Inverse and Determinant    ####
  ####################################################################
  ####       version 1.0 Created by Chong Liu on 2012/11/04       ####
  ####################################################################
  #
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#                                              
  # Input
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.MatSym:         A matrix of character strings, which stores the structure of matrix with each entry denoted by charater names. 
  #                   Variables names start with capital letter A followed by number index. Variable names are arranged from left to right by rows
  #                   and number indices are in ascending order.
  # 2.InvSym:         If yacas objects of "matrix","inverse","determinant" are ready from outside, they can be provided for evaluation directly
  # 3.Entry:          A matrix of size NEval by NEntry, with each row being the value of all entries in the matrix, aligned by row. 
  #                   For example, matrix |A1,A2| is stored in a row by (A1,A2,A3).
  #                                       |A3,A3| 
  # 4.Evaluate:       A logical value indicating whether inverse will be numerically evaluated. 
  # 5.Output:         A vector of item names to be returned; 
  #                   "Matrix" is matrices themselves;
  #                   "Inverse" is inverse of all matrices;
  #                   "Determinant" is determinant of all matrices.
  # 6.MatrixFold:     Logical value indicating whether Matrix is of size NEval by NRow*NCol (False) or NRow by NCol by NEval (True). 
  # 7.InverseFold:    Logical value indicating whether Inverse is of size NEval by NRow*NCol (False) or NRow by NCol by NEval (True).
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # Output
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  # 1.Matrix          A yacas object, or A array of orginal matrices of size either NEval by NRow*NCol (MatrixFolder = False) or NRow by NCol by NEval (MatrixFold = True), or NULL (not returned).
  # 2.Inverse         A yacas object, or A array of inverse matrices of size either NEval by NRow*NCol (MatrixFolder = False) or NRow by NCol by NEval (MatrixFold = True), or NULL (not returned).
  # 3.Determinant     A yacas object, or A vector of length NEval of determinant of original matrices.
  #-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------#
  
  InvMat <- function(MatSym, InvSym = NULL, Entry = NULL, Evaluate = TRUE, Output = "Inverse", MatrixFold = TRUE, InverseFold = TRUE)
  {
    # include packages and functions
    
      library(Ryacas) # load Ryacas package
      #yacasInstall() # install the package
    
    # check consistency
      
      if(is.null(Entry) & Evaluate){stop("Entry must be provided if matrix inverse is to be numerically evaluated")}
      if(is.null(MatSym) & is.null(InvSym)){stop("Either MatSym or InvSym must be provided")}
      
    # extract variable names for matrix entries
      
      NCol <- NRow <- dim(MatSym)[1]
      EntryNames <- unique(as.vector(t(MatSym))) # number index should be in ascending order
      NEntry <- length(EntryNames)

    # generate character string for yacas matrix construction
      
      if(is.null(InvSym))
      {
        String <- "{"
        for(r in 1 : NRow)
        {
          String <- paste(String, "{", paste(MatSym[r,], collapse = ","), "}", sep = "")
          if(r < NRow){String <- paste(String, ",", sep = "")}
        }
        String <- paste("A:=", String, "}", sep = "")
      }
      
    # construct yacas matrix, yacas inverse and yacas determinant

      if(is.null(InvSym))
      {        
        A <- yacas(eval(String))      
        AInv <- yacas("AInv:=Inverse(A)")      
        ADet <- yacas("ADet:=Determinant(A)")
        #print("no InvSym :(")
      }else
      {
        A <- InvSym$Matrix
        AInv <- InvSym$Inverse
        ADet <- InvSym$Determinant
        #print("InvSym available!")
      }
      
    # evaluate for output
      
      if(Evaluate)
      {
        NEval <- dim(Entry)[1] # number of matrix evaluations
        Output <- tolower(Output) # transfer to lowercase
        StringEval <- paste("list(", paste(EntryNames, " = Entry[,", 1 : NEntry, "]", sep = "", collapse = ", "), ")", sep = "") # character string used to evaluate matrix, inverse or determinant
                                           
        if(is.element("matrix", Output))
        {
          Temp <- Eval(A, eval(parse(text = StringEval)))
          if(MatrixFold){Matrix <- array(unlist(Temp), dim = c(NEval, NCol, NRow)); Matrix <- aperm(Matrix, perm = c(3,2,1))}else
          {Matrix <- array(unlist(Temp), dim = c(NEval, NRow*NCol))}
        }else{Matrix <- NULL}
                                           
        if(is.element("inverse", Output))
        {
          Temp <- Eval(AInv, eval(parse(text = StringEval)))
          if(InverseFold){Inverse <- array(unlist(Temp), dim = c(NEval, NCol, NRow)); Inverse <- aperm(Inverse, perm = c(3,2,1))}else
          {Inverse <- array(unlist(Temp), dim = c(NEval, NRow*NCol))}
        }else{Inverse <- NULL}
                                           
        if(is.element("determinant", Output))
        {
          Temp <- Eval(ADet, eval(parse(text = StringEval)))
          Determinant <- unlist(Temp)
        }else{Determinant <- NULL}  
      }
      
    # return results
      
      if(Evaluate){results <- list(Matrix = Matrix, Inverse = Inverse, Determinant = Determinant)}else
      {results <- list(Matrix = A, Inverse = AInv, Determinant = ADet)}
      
      return(results)      
    }